from .bqplot import PlotBase
import ipyvolume.pylab as p3
import ipyvolume.examples
import traitlets
import vaex.dataset
import ipywidgets as widgets
from IPython.display import HTML, display_html, display_javascript, display
import numpy as np

class PlotDefault(PlotBase):
    z = traitlets.Unicode(allow_none=False)

    def _update_image(self):
        with self.output:
            grid = self.get_grid()
            if self.smooth_pre:
                for i in range(grid.shape[0]):
                    grid[i] = vaex.grids.gf(grid[i], self.smooth_pre)
            f = vaex.dataset._parse_f(self.f)
            fgrid = f(grid)
            if self.smooth_post:
                for i in range(grid.shape[0]):
                    fgrid[i] = vaex.grids.gf(fgrid[i], self.smooth_post)
            ngrid, fmin, fmax = self.normalise(fgrid)
            print(ngrid.shape)
            if len(ngrid.shape) == 4:
                #if ngrid.shape[0] == 1:
                ngrid = ngrid[-1]
            p3.volshow(ngrid.T, controls=self._first_time)

            vx, vy, vz = self.vgrids[:3]
            vcount = self.vcount
            if vx is not None and vy is not None and vz is not None and vcount is not None:
                vcount = vcount[-1] # no multivolume render, just take the last selection
                vx = vx[-1]
                vy = vy[-1]
                vz = vz[-1]
                print(vx.shape)
                ok = np.isfinite(vx) & np.isfinite(vy) & np.isfinite(vz)
                vcount_min = None
                vcount_max = None
                if self.vcount_limits is not None:
                    try:
                        vcount_min, vcount_max = self.vcount_limits
                    except:
                        vcount_min = self.vcount_limits
                if vcount_min is not None:
                    ok &= (vcount > vcount_min)
                if vcount_max is not None:
                    ok &= (vcount < vcount_max)
                x, y, z = ipyvolume.examples.xyz(self.get_vshape()[0], limits=self.limits, sparse=False, centers=True)
                v1d = [k[ok] for k in [x, y, z, vx, vy, vz]]
                vsize = 5
                vcolor = "grey"
                if self._first_time:
                    self.quiver = p3.quiver(*v1d, size=vsize, color=vcolor)
                else:
                    self.quiver.x = x[ok]
                    self.quiver.y = y[ok]
                    self.quiver.z = z[ok]
                    self.quiver.vx = vx[ok]
                    self.quiver.vy = vy[ok]
                    self.quiver.vz = vz[ok]

            p3.xlim(*self.limits[0])
            p3.ylim(*self.limits[1])
            p3.zlim(*self.limits[2])
            self._first_time = False

    def create_plot(self):
        self._first_time = True
        self.figure = p3.figure()

    def show(self):
        container = p3.gcc()
        vbox = widgets.VBox([container, self.progress, widgets.VBox(self.tools), self.output])
        display(vbox)

    def create_tools(self):
        self.tools = []

        callback = self.dataset.signal_selection_changed.connect(lambda *x: self.update_grid())

        def cleanup(callback=callback):
            self.dataset.signal_selection_changed.disconnect(callback=callback)

        self._cleanups.append(cleanup)

    def get_binby(self):
        return [self.x, self.y, self.z]


__author__ = 'maartenbreddels'
import aplus
import os

"""
Mini library to make working with promises/futures a bit more Python like.

Example:

@delayed
def f(grid):
    return grid**2

f(ds.count(delay=True))

"""


def promisify(value):
    # TODO, support futures etc..
    if isinstance(value, aplus.Promise):
        return value
    if isinstance(value, (list, tuple)):
        return aplus.listPromise(*list([promisify(k) for k in value]))
    else:
        return aplus.Promise.fulfilled(value)


def delayed(f):
    '''Decorator to transparantly accept delayed computation.

    Example:

    >>> delayed_sum = ds.sum(ds.E, binby=ds.x, limits=limits,
    >>>                   shape=4, delay=True)
    >>> @vaex.delayed
    >>> def total_sum(sums):
    >>>     return sums.sum()
    >>> sum_of_sums = total_sum(delayed_sum)
    >>> ds.execute()
    >>> sum_of_sums.get()
    See the tutorial for a more complete example https://docs.vaex.io/en/latest/tutorial.html#Parallel-computations
    '''

    def wrapped(*args, **kwargs):
        # print "calling", f, "with", kwargs
        # key_values = kwargs.items()
        key_promise = list([(key, promisify(value)) for key, value in kwargs.items()])
        # key_promise = [(key, promisify(value)) for key, value in key_values]
        arg_promises = list([promisify(value) for value in args])
        kwarg_promises = list([promise for key, promise in key_promise])
        promises = arg_promises + kwarg_promises
        for promise in promises:
            def echo_error(exc, promise=promise):
                print("error with ", promise, "exception is", exc)
                # raise exc

            def echo(value, promise=promise):
                print("done with ", repr(promise), "value is", value)
            # promise.then(echo, echo_error)

        # print promises
        allarguments = aplus.listPromise(*promises)

        def call(_):
            kwargs_real = {key: promise.get() for key, promise in key_promise}
            args_real = list([promise.get() for promise in arg_promises])
            return f(*args_real, **kwargs_real)

        def error(exc):
            if os.environ.get('VAEX_DEBUG', False):
                print("Error from delayed", exc)
            # import vaex
            # vaex.utils.print_stack_trace()
            raise exc
        return allarguments.then(call, error)
    return wrapped


@delayed
def delayed_args(*args):
    return args


@delayed
def delayed_list(l):
    return delayed_args(*l)



import numpy as np
import pyarrow as pa
import pyarrow.compute as pc
import vaex
from ..expression import _binary_ops, _unary_ops, reversable
from .utils import combine_missing

class NumpyDispatch:
    def __init__(self, ar):
        self._array = ar
        if isinstance(ar, vaex.column.ColumnStringArrow):
            ar = pa.array(ar)
        if isinstance(ar, np.ndarray):
            self._numpy_array = ar
            self._arrow_array = None
        elif isinstance(ar, vaex.array_types.supported_arrow_array_types):
            self._numpy_array = None
            self._arrow_array = ar
        else:
            raise TypeError(f'Only support numpy and arrow, not {type(ar)}')

    def add_missing(self, ar):
        if isinstance(ar, np.ndarray):
            # if we are an arrow array, we upgrade ar to one
            if isinstance(self._array, vaex.array_types.supported_arrow_array_types):
                ar = vaex.array_types.to_arrow(ar)
                ar = combine_missing(ar, self._array)
            # else: both numpy, handled by numpy
        else:
            if isinstance(self._array, vaex.array_types.supported_arrow_array_types):
                ar = combine_missing(ar, self._array)
            # else: was numpy, handled by numpy
        return ar


    @property
    def numpy_array(self):
        if self._numpy_array is None:
            import vaex.arrow.convert
            arrow_array = self._arrow_array
            arrow_array = vaex.arrow.convert.ensure_not_chunked(arrow_array)
            buffers = arrow_array.buffers()
            # for math, we don't care about the nulls
            if buffers[0] is not None:
                buffers[0] = None
                arrow_array = pa.Array.from_buffers(arrow_array.type, len(arrow_array), buffers, offset=arrow_array.offset)
            self._numpy_array = vaex.array_types.to_numpy(arrow_array)
        return self._numpy_array

    @property
    def arrow_array(self):
        if self._arrow_array is None:
            # convert lazily, since not all arrow arrays (e.g. lists) can be converted
            if self._arrow_array is None:
                self._arrow_array = vaex.array_types.to_arrow(self._numpy_array)
        return self._arrow_array


for op in _binary_ops:
    def closure(op=op):
        def operator(a, b):
            a_data = a
            b_data = b
            if isinstance(a, NumpyDispatch):
                a_data = a.numpy_array
            if isinstance(b, NumpyDispatch):
                b_data = b.numpy_array
            if op['name'] == 'eq' and (vaex.array_types.is_string(a_data) or vaex.array_types.is_string(b_data)):
                result_data = vaex.functions.str_equals(a_data, b_data)
            else:
                result_data = op['op'](a_data, b_data)
            if isinstance(a, NumpyDispatch):
                result_data = a.add_missing(result_data)
            if isinstance(b, NumpyDispatch):
                result_data = b.add_missing(result_data)
            return NumpyDispatch(result_data)
        return operator
    method_name = '__%s__' % op['name']
    setattr(NumpyDispatch, method_name, closure())
     # to support e.g. (1 + ...)    # to support e.g. (1 + ...)
    if op['name'] in reversable:
        def closure(op=op):
            def operator(b, a):
                a_data = a
                b_data = b
                if isinstance(a, NumpyDispatch):
                    a_data = a.numpy_array
                if isinstance(b, NumpyDispatch):
                    b_data = b.numpy_array
                result_data = op['op'](a_data, b_data)
                if isinstance(a, NumpyDispatch):
                    result_data = a.add_missing(result_data)
                if isinstance(b, NumpyDispatch):
                    result_data = b.add_missing(result_data)
                return NumpyDispatch(result_data)
            return operator
        method_name = '__r%s__' % op['name']
        setattr(NumpyDispatch, method_name, closure())


for op in _unary_ops:
    def closure(op=op):
        def operator(a):
            a_data = a.numpy_array
            result_data = op['op'](a_data)
            if isinstance(a, NumpyDispatch):
                result_data = a.add_missing(result_data)
            return NumpyDispatch(result_data)
        return operator
    method_name = '__%s__' % op['name']
    setattr(NumpyDispatch, method_name, closure())


def wrap(value):
    if not isinstance(value, NumpyDispatch): # and not isinstance(value, np.ndarray):
        if isinstance(value, vaex.array_types.supported_array_types + (vaex.column.ColumnStringArrow,)):
            return NumpyDispatch(value)
    # for performance reasons we don't visit lists and dicts
    return value


def unwrap(value):
    if isinstance(value, NumpyDispatch):
        return value._array
    # for performance reasons we don't visit lists and dicts
    return value


def autowrapper(f):
    '''Takes a function f, and will unwrap all its arguments and wrap the return value'''
    def wrapper(*args, **kwargs):
        args_original = args
        args = list(map(unwrap, args))
        kwargs = {k: unwrap(v) for k, v, in kwargs.items()}
        result = f(*args, **kwargs)
        return wrap(result)
    return wrapper



import ast
import os
import base64
import cloudpickle as pickle
import functools
import operator
import six
import collections
import weakref

from future.utils import with_metaclass
import numpy as np
import tabulate
import pyarrow as pa
from vaex.datatype import DataType
from vaex.docstrings import docsubst

from vaex.utils import _ensure_strings_from_expressions, _ensure_string_from_expression
from vaex.column import ColumnString, _to_string_sequence
from .hash import counter_type_from_dtype
import vaex.serialize
from . import expresso


try:
    from StringIO import StringIO
except ImportError:
    from io import BytesIO as StringIO

try:
    collectionsAbc = collections.abc
except AttributeError:
    collectionsAbc = collections


# TODO: repeated from dataframe.py
default_shape = 128
PRINT_MAX_COUNT = 10

expression_namespace = {}
expression_namespace['nan'] = np.nan


expression_namespace = {}
expression_namespace['nan'] = np.nan


_binary_ops = [
    dict(code="+", name='add', op=operator.add),
    dict(code="in", name='contains', op=operator.contains),
    dict(code="/", name='truediv', op=operator.truediv),
    dict(code="//", name='floordiv', op=operator.floordiv),
    dict(code="&", name='and', op=operator.and_),
    dict(code="^", name='xor', op=operator.xor),

    dict(code="|", name='or', op=operator.or_),
    dict(code="**", name='pow', op=operator.pow),
    dict(code="is", name='is', op=operator.is_),
    dict(code="is not", name='is_not', op=operator.is_not),

    dict(code="<<", name='lshift', op=operator.lshift),
    dict(code="%", name='mod', op=operator.mod),
    dict(code="*", name='mul', op=operator.mul),

    dict(code=">>", name='rshift', op=operator.rshift),
    dict(code="-", name='sub', op=operator.sub),

    dict(code="<", name='lt', op=operator.lt),
    dict(code="<=", name='le', op=operator.le),
    dict(code="==", name='eq', op=operator.eq),
    dict(code="!=", name='ne', op=operator.ne),
    dict(code=">=", name='ge', op=operator.ge),
    dict(code=">", name='gt', op=operator.gt),
]
if hasattr(operator, 'div'):
    _binary_ops.append(dict(code="/", name='div', op=operator.div))
if hasattr(operator, 'matmul'):
    _binary_ops.append(dict(code="@", name='matmul', op=operator.matmul))

reversable = 'add sub mul matmul truediv floordiv mod divmod pow lshift rshift and xor or'.split()

_unary_ops = [
    dict(code="~", name='invert', op=operator.invert),
    dict(code="-", name='neg', op=operator.neg),
    dict(code="+", name='pos', op=operator.pos),
]


class Meta(type):
    def __new__(upperattr_metaclass, future_class_name,
                future_class_parents, attrs):
        # attrs = {}
        for op in _binary_ops:
            def wrap(op=op):
                def f(a, b):
                    self = a
                    # print(op, a, b)
                    try:
                        stringy = isinstance(b, str) or b.is_string()
                    except:
                        # this can happen when expression is a literal, like '1' (used in propagate_unc)
                        # which causes the dtype to fail
                        stringy = False
                    if stringy:
                        if isinstance(b, str):
                            b = repr(b)
                        if op['code'] == '==':
                            expression = 'str_equals({0}, {1})'.format(a.expression, b)
                        elif op['code'] == '+':
                            expression = 'str_cat({0}, {1})'.format(a.expression, b)
                        else:
                            raise ValueError('operand %r not supported for string comparison' % op['code'])
                        return Expression(self.ds, expression=expression)
                    else:
                        if isinstance(b, Expression):
                            assert b.ds == a.ds
                            b = b.expression
                        elif isinstance(b, (np.timedelta64)):
                            df = a.ds
                            b = df.add_variable('var_time_delta', b, unique=True)
                        elif isinstance(b, (np.datetime64)):
                            df = a.ds
                            b = df.add_variable('var_date_time', b, unique=True)
                        expression = '({0} {1} {2})'.format(a.expression, op['code'], b)
                    return Expression(self.ds, expression=expression)
                attrs['__%s__' % op['name']] = f
                if op['name'] in reversable:
                    def f(a, b):
                        self = a
                        if isinstance(b, str):
                            if op['code'] == '+':
                                expression = 'str_cat({1}, {0})'.format(a.expression, repr(b))
                            else:
                                raise ValueError('operand %r not supported for string comparison' % op['code'])
                            return Expression(self.ds, expression=expression)
                        else:
                            if isinstance(b, Expression):
                                assert b.ds == a.ds
                                b = b.expression
                            expression = '({2} {1} {0})'.format(a.expression, op['code'], b)
                            return Expression(self.ds, expression=expression)
                    attrs['__r%s__' % op['name']] = f

            wrap(op)
        for op in _unary_ops:
            def wrap(op=op):
                def f(a):
                    self = a
                    expression = '{0}({1})'.format(op['code'], a.expression)
                    return Expression(self.ds, expression=expression)
                attrs['__%s__' % op['name']] = f
            wrap(op)
        return type(future_class_name, future_class_parents, attrs)


class DateTime(object):
    """DateTime operations

    Usually accessed using e.g. `df.birthday.dt.dayofweek`
    """
    def __init__(self, expression):
        self.expression = expression


class TimeDelta(object):
    """TimeDelta operations

    Usually accessed using e.g. `df.delay.td.days`
    """
    def __init__(self, expression):
        self.expression = expression


class StringOperations(object):
    """String operations.

    Usually accessed using e.g. `df.name.str.lower()`
    """
    def __init__(self, expression):
        self.expression = expression


class StringOperationsPandas(object):
    """String operations using Pandas Series (much slower)"""
    def __init__(self, expression):
        self.expression = expression


class Expression(with_metaclass(Meta)):
    """Expression class"""
    def __init__(self, ds, expression, ast=None):
        self.ds = ds
        assert not isinstance(ds, Expression)
        if isinstance(expression, Expression):
            expression = expression.expression
        self._ast = ast
        self._expression = expression
        self.df._expressions.append(weakref.ref(self))
        self._ast_names = None

    def copy(self, df=None):
        """Efficiently copies an expression.

        Expression objects have both a string and AST representation. Creating
        the AST representation involves parsing the expression, which is expensive.

        Using copy will deepcopy the AST when the expression was already parsed.

        :param df: DataFrame for which the expression will be evaluated (self.df if None)
        """
        # expression either has _expression or _ast not None
        if df is None:
            df = self.df
        if self._expression is not None:
            expression = Expression(df, self._expression)
            if self._ast is not None:
                expression._ast = copy.deepcopy(self._ast)
        elif self._ast is not None:
            expression = Expression(df, copy.deepcopy(self._ast))
            if self._ast is not None:
                expression._ast = self._ast
        return expression

    @property
    def ast(self):
        """Returns the abstract syntax tree (AST) of the expression"""
        if self._ast is None:
            self._ast = expresso.parse_expression(self.expression)
        return self._ast

    @property
    def ast_names(self):
        if self._ast_names is None:
            self._ast_names = expresso.names(self.ast)
        return self._ast_names

    @property
    def _ast_slices(self):
        return expresso.slices(self.ast)

    @property
    def expression(self):
        if self._expression is None:
            self._expression = expresso.node_to_string(self.ast)
        return self._expression

    @expression.setter
    def expression(self, value):
        # if we reassign to expression, we clear the ast cache
        if value != self._expression:
            self._expression = value
            self._ast = None

    def __bool__(self):
        """Cast expression to boolean. Only supports (<expr1> == <expr2> and <expr1> != <expr2>)

        The main use case for this is to support assigning to traitlets. e.g.:

        >>> bool(expr1 == expr2)

        This will return True when expr1 and expr2 are exactly the same (in string representation). And similarly for:

        >>> bool(expr != expr2)

        All other cases will return True.
        """
        # this is to make traitlets detect changes
        import _ast
        if isinstance(self.ast, _ast.Compare) and len(self.ast.ops) == 1 and isinstance(self.ast.ops[0], _ast.Eq):
            return expresso.node_to_string(self.ast.left) == expresso.node_to_string(self.ast.comparators[0])
        if isinstance(self.ast, _ast.Compare) and len(self.ast.ops) == 1 and isinstance(self.ast.ops[0], _ast.NotEq):
            return expresso.node_to_string(self.ast.left) != expresso.node_to_string(self.ast.comparators[0])
        return True


    @property
    def df(self):
        # lets gradually move to using .df
        return self.ds

    @property
    def dtype(self):
        return self.df.data_type(self.expression)

    # TODO: remove this method?
    def data_type(self, array_type=None, axis=0):
        return self.df.data_type(self.expression, axis=axis)

    @property
    def shape(self):
        return self.df._shape_of(self)

    def to_arrow(self, convert_to_native=False):
        '''Convert to Apache Arrow array (will byteswap/copy if convert_to_native=True).'''
        values = self.values
        return vaex.array_types.to_arrow(values, convert_to_native=convert_to_native)

    def __arrow_array__(self, type=None):
        values = self.to_arrow()
        return pa.array(values, type=type)

    def to_numpy(self, strict=True):
        """Return a numpy representation of the data"""
        values = self.values
        return vaex.array_types.to_numpy(values, strict=strict)

    def to_dask_array(self, chunks="auto"):
        import dask.array as da
        import uuid
        dtype = self.dtype
        chunks = da.core.normalize_chunks(chunks, shape=self.shape, dtype=dtype.numpy)
        name = 'vaex-expression-%s' % str(uuid.uuid1())
        def getitem(df, item):
            assert len(item) == 1
            item = item[0]
            start, stop, step = item.start, item.stop, item.step
            assert step in [None, 1]
            return self.evaluate(start, stop, parallel=False)
        dsk = da.core.getem(name, chunks, getitem=getitem, shape=self.shape, dtype=dtype.numpy)
        dsk[name] = self
        return da.Array(dsk, name, chunks, dtype=dtype.numpy)

    def to_pandas_series(self):
        """Return a pandas.Series representation of the expression.

        Note: Pandas is likely to make a memory copy of the data.
        """
        import pandas as pd
        return pd.Series(self.values)

    def __getitem__(self, slice):
        return self.ds[slice][self.expression]

    def __abs__(self):
        """Returns the absolute value of the expression"""
        return self.abs()

    @property
    def dt(self):
        """Gives access to datetime operations via :py:class:`DateTime`"""
        return DateTime(self)

    @property
    def td(self):
        """Gives access to timedelta operations via :py:class:`TimeDelta`"""
        return TimeDelta(self)

    @property
    def str(self):
        """Gives access to string operations via :py:class:`StringOperations`"""
        return StringOperations(self)

    @property
    def str_pandas(self):
        """Gives access to string operations via :py:class:`StringOperationsPandas` (using Pandas Series)"""
        return StringOperationsPandas(self)

    @property
    def values(self):
        return self.evaluate()

    def derivative(self, var, simplify=True):
        var = _ensure_string_from_expression(var)
        return self.__class__(self.ds, expresso.derivative(self.ast, var, simplify=simplify))

    def expand(self, stop=[]):
        """Expand the expression such that no virtual columns occurs, only normal columns.

        Example:

        >>> df = vaex.example()
        >>> r = np.sqrt(df.data.x**2 + df.data.y**2)
        >>> r.expand().expression
        'sqrt(((x ** 2) + (y ** 2)))'

        """
        stop = _ensure_strings_from_expressions(stop)
        def translate(id):
            if id in self.ds.virtual_columns and id not in stop:
                return self.ds.virtual_columns[id]
        expr = expresso.translate(self.ast, translate)
        return Expression(self.ds, expr)

    def variables(self, ourself=False, expand_virtual=True, include_virtual=True):
        """Return a set of variables this expression depends on.

        Example:

        >>> df = vaex.example()
        >>> r = np.sqrt(df.data.x**2 + df.data.y**2)
        >>> r.variables()
        {'x', 'y'}
        """
        variables = set()
        def record(varname):
            # do this recursively for virtual columns
            if varname in self.ds.virtual_columns and varname not in variables:
                if (include_virtual and (varname != self.expression)) or (varname == self.expression and ourself):
                    variables.add(varname)
                if expand_virtual:
                    variables.update(self.df[self.df.virtual_columns[varname]].variables(ourself=include_virtual, include_virtual=include_virtual))
            # we usually don't want to record ourself
            elif varname != self.expression or ourself:
                variables.add(varname)

        expresso.translate(self.ast, record)
        # df is a buildin, don't record it, if df is a column name, it will be collected as
        # df['df']
        variables -= {'df'}
        for varname in self._ast_slices:
            if varname in self.df.virtual_columns and varname not in variables:
                if (include_virtual and (f"df['{varname}']" != self.expression)) or (f"df['{varname}']" == self.expression and ourself):
                    variables.add(varname)
                if expand_virtual:
                    if varname in self.df.virtual_columns:
                        variables |= self.df[self.df.virtual_columns[varname]].variables(ourself=include_virtual, include_virtual=include_virtual)
            elif f"df['{varname}']" != self.expression or ourself:
                variables.add(varname)

        return variables

    def _graph(self):
        """"Return a graph containing the dependencies of this expression
        Structure is:
            [<string expression>, <function name if callable>, <function object if callable>, [subgraph/dependencies, ....]]
        """
        expression = self.expression

        def walk(node):
            if isinstance(node, six.string_types):
                if node in self.ds.virtual_columns:
                    ex = Expression(self.ds, self.ds.virtual_columns[node])
                    return [node, None, None, [ex._graph()]]
                else:
                    return node
            else:
                fname, node_repr, deps = node
                if len(node_repr) > 30:  # clip too long expressions
                    node_repr = node_repr[:26] + ' ....'
                deps = [walk(dep) for dep in deps]
                obj = self.ds.functions.get(fname)
                # we don't want the wrapper, we want the underlying object
                if isinstance(obj, Function):
                    obj = obj.f
                if isinstance(obj, FunctionSerializablePickle):
                    obj = obj.f
                return [node_repr, fname, obj, deps]
        return walk(expresso._graph(expression))

    def _graphviz(self, dot=None):
        """Return a graphviz.Digraph object with a graph of the expression"""
        from graphviz import Graph, Digraph
        node = self._graph()
        dot = dot or Digraph(comment=self.expression)
        def walk(node):
            if isinstance(node, six.string_types):
                dot.node(node, node)
                return node, node
            else:
                node_repr, fname, fobj, deps = node
                node_id = node_repr
                dot.node(node_id, node_repr)
                for dep in deps:
                    dep_id, dep = walk(dep)
                    dot.edge(node_id, dep_id)
                return node_id, node
        walk(node)
        return dot

    def __str__(self):
        return self.expression

    # def __array__(self, dtype=None):
    #     '''For casting to a numpy array

    #     Example:
    #         >>> np.array(ds.x**2)

    #     '''
    #     return self.ds.evaluate(self)

    def tolist(self, i1=None, i2=None):
        '''Short for expr.evaluate().tolist()'''
        values = self.evaluate(i1=i1, i2=i2)
        if isinstance(values, (pa.Array, pa.ChunkedArray)):
            return values.to_pylist()
        return values.tolist()

    if not os.environ.get('VAEX_DEBUG', ''):
        def __repr__(self):
            return self._repr_plain_()

    def _repr_plain_(self):
        from .formatting import _format_value
        def format(values):
            for i in range(len(values)):
                value = values[i]
                yield _format_value(value)
        colalign = ("right",) * 2
        try:
            N = len(self.ds)
            if N <= PRINT_MAX_COUNT:
                values = format(self.evaluate(0, N))
                values = tabulate.tabulate([[i, k] for i, k in enumerate(values)], tablefmt='plain', colalign=colalign)
            else:
                values_head = format(self.evaluate(0, PRINT_MAX_COUNT//2))
                values_tail = format(self.evaluate(N - PRINT_MAX_COUNT//2, N))
                values_head = list(zip(range(PRINT_MAX_COUNT//2), values_head)) +\
                              list(zip(range(N - PRINT_MAX_COUNT//2, N), values_tail))
                values = tabulate.tabulate([k for k in values_head], tablefmt='plain', colalign=colalign)
                values = values.split('\n')
                width = max(map(len, values))
                separator = '\n' + '...'.center(width, ' ') + '\n'
                values = "\n".join(values[:PRINT_MAX_COUNT//2]) + separator + "\n".join(values[PRINT_MAX_COUNT//2:]) + '\n'
        except Exception as e:
            values = 'Error evaluating: %r' % e
        expression = self.expression
        if len(expression) > 60:
            expression = expression[:57] + '...'
        info = 'Expression = ' + expression + '\n'
        dtype = self.dtype
        if self.expression in self.ds.get_column_names(hidden=True):
            state = "column"
        elif self.expression in self.ds.get_column_names(hidden=True):
            state = "virtual column"
        else:
            state = "expression"
        line = 'Length: {:,} dtype: {} ({})\n'.format(len(self.ds), dtype, state)
        info += line
        info += '-' * (len(line)-1) + '\n'
        info += values
        return info

    def count(self, binby=[], limits=None, shape=default_shape, selection=False, delay=False, edges=False, progress=None):
        '''Shortcut for ds.count(expression, ...), see `Dataset.count`'''
        kwargs = dict(locals())
        del kwargs['self']
        kwargs['expression'] = self.expression
        return self.ds.count(**kwargs)

    def sum(self, axis=None, binby=[], limits=None, shape=default_shape, selection=False, delay=False, progress=None):
        '''Sum elements over given axis.

        If no axis is given, it will sum over all axes.

        For non list elements, this is a shortcut for ds.sum(expression, ...), see `Dataset.sum`.

        >>> list_data = [1, 2, None], None, [], [1, 3, 4, 5]
        >>> df = vaex.from_arrays(some_list=pa.array(list_data))
        >>> df.some_list.sum().item()  # will sum over all axis
        16
        >>> df.some_list.sum(axis=1).tolist()  # sums the list elements
        [3, None, 0, 13]

        :param int axis: Axis over which to determine the unique elements (None will flatten arrays or lists)
        '''
        expression = self
        if axis is None:
            axis = [0]
            dtype = self.dtype
            while dtype.is_list:
                axis.append(axis[-1] + 1)
                dtype = dtype.value_type
        elif not isinstance(axis, list):
            axis = [axis]
            axis = list(set(axis))  # remove repeated elements
        dtype = self.dtype
        if 1 in axis:
            if self.dtype.is_list:
                expression = expression.list_sum()
                if axis:
                    axis.remove(1)
            else:
                raise ValueError(f'axis=1 not supported for dtype={dtype}')
        if axis and axis[0] != 0:
            raise ValueError(f'Only axis 0 or 1 is supported')
        if axis is None or 0 in axis:
            kwargs = dict(locals())
            del kwargs['self']
            del kwargs['axis']
            del kwargs['dtype']
            kwargs['expression'] = expression.expression
            return self.ds.sum(**kwargs)
        else:
            return expression

    def mean(self, binby=[], limits=None, shape=default_shape, selection=False, delay=False, progress=None):
        '''Shortcut for ds.mean(expression, ...), see `Dataset.mean`'''
        kwargs = dict(locals())
        del kwargs['self']
        kwargs['expression'] = self.expression
        return self.ds.mean(**kwargs)

    def std(self, binby=[], limits=None, shape=default_shape, selection=False, delay=False, progress=None):
        '''Shortcut for ds.std(expression, ...), see `Dataset.std`'''
        kwargs = dict(locals())
        del kwargs['self']
        kwargs['expression'] = self.expression
        return self.ds.std(**kwargs)

    def var(self, binby=[], limits=None, shape=default_shape, selection=False, delay=False, progress=None):
        '''Shortcut for ds.std(expression, ...), see `Dataset.var`'''
        kwargs = dict(locals())
        del kwargs['self']
        kwargs['expression'] = self.expression
        return self.ds.var(**kwargs)

    def minmax(self, binby=[], limits=None, shape=default_shape, selection=False, delay=False, progress=None):
        '''Shortcut for ds.minmax(expression, ...), see `Dataset.minmax`'''
        kwargs = dict(locals())
        del kwargs['self']
        kwargs['expression'] = self.expression
        return self.ds.minmax(**kwargs)

    def min(self, binby=[], limits=None, shape=default_shape, selection=False, delay=False, progress=None):
        '''Shortcut for ds.min(expression, ...), see `Dataset.min`'''
        kwargs = dict(locals())
        del kwargs['self']
        kwargs['expression'] = self.expression
        return self.ds.min(**kwargs)

    def max(self, binby=[], limits=None, shape=default_shape, selection=False, delay=False, progress=None):
        '''Shortcut for ds.max(expression, ...), see `Dataset.max`'''
        kwargs = dict(locals())
        del kwargs['self']
        kwargs['expression'] = self.expression
        return self.ds.max(**kwargs)

    def nop(self):
        """Evaluates expression, and drop the result, usefull for benchmarking, since vaex is usually lazy"""
        return self.ds.nop(self.expression)

    @property
    def transient(self):
        """If this expression is not transient (e.g. on disk) optimizations can be made"""
        return self.expand().expression not in self.ds.columns

    @property
    def masked(self):
        """Alias to df.is_masked(expression)"""
        return self.ds.is_masked(self.expression)

    def value_counts(self, dropna=False, dropnan=False, dropmissing=False, ascending=False, progress=False, axis=None):
        """Computes counts of unique values.

         WARNING:
          * If the expression/column is not categorical, it will be converted on the fly
          * dropna is False by default, it is True by default in pandas

        :param dropna: when True, it will not report the NA (see :func:`Expression.isna`)
        :param dropnan: when True, it will not report the nans(see :func:`Expression.isnan`)
        :param dropmissing: when True, it will not report the missing values (see :func:`Expression.ismissing`)
        :param ascending: when False (default) it will report the most frequent occuring item first
        :param bool axis: Axis over which to determine the unique elements (None will flatten arrays or lists)
        :returns: Pandas series containing the counts
        """
        from pandas import Series
        if axis is not None:
            raise ValueError('only axis=None is supported')

        data_type = self.data_type()
        data_type_item = self.data_type(axis=-1)

        transient = self.transient or self.ds.filtered or self.ds.is_masked(self.expression)
        if self.is_string() and not transient:
            # string is a special case, only ColumnString are not transient
            ar = self.ds.columns[self.expression]
            if not isinstance(ar, ColumnString):
                transient = True

        counter_type = counter_type_from_dtype(data_type_item, transient)
        counters = [None] * self.ds.executor.thread_pool.nthreads
        def map(thread_index, i1, i2, ar):
            if counters[thread_index] is None:
                counters[thread_index] = counter_type()
            if data_type.is_list and axis is None:
                ar = ar.values
            if data_type_item.is_string:
                ar = _to_string_sequence(ar)
            else:
                ar = vaex.array_types.to_numpy(ar)
            if np.ma.isMaskedArray(ar):
                mask = np.ma.getmaskarray(ar)
                counters[thread_index].update(ar, mask)
            else:
                counters[thread_index].update(ar)
            return 0
        def reduce(a, b):
            return a+b
        self.ds.map_reduce(map, reduce, [self.expression], delay=False, progress=progress, name='value_counts', info=True, to_numpy=False)
        counters = [k for k in counters if k is not None]
        counter0 = counters[0]
        for other in counters[1:]:
            counter0.merge(other)
        value_counts = counter0.extract()
        index = np.array(list(value_counts.keys()))
        counts = np.array(list(value_counts.values()))

        order = np.argsort(counts)
        if not ascending:
            order = order[::-1]
        counts = counts[order]
        index = index[order]
        # nan can already be present for dtype=object, remove it
        nan_mask = index != index
        if np.any(nan_mask):
            index = index[~mask]
            counts = index[~mask]
        # nan can already be present for dtype=object, optionally remove it
        none_mask = index == None
        if np.any(none_mask):
            index = index.tolist()
            counts = counts.tolist()
            i = index.index(None)
            if (dropmissing or dropna):
                del index[i]
                del counts[i]
            else:
                index[i] = "missing"
            index = np.array(index)
            counts = np.array(counts)

        if not dropna or not dropnan or not dropmissing:
            index = index.tolist()
            counts = counts.tolist()
            if not (dropnan or dropna) and counter0.nan_count:
                index = [np.nan] + index
                counts = [counter0.nan_count] + counts
            if not (dropmissing or dropna) and counter0.null_count:
                index = ['missing'] + index
                counts = [counter0.null_count] + counts

        return Series(counts, index=index)

    @docsubst
    def unique(self, dropna=False, dropnan=False, dropmissing=False, selection=None, axis=None, array_type='list', delay=False):
        """Returns all unique values.

        :param dropmissing: do not count missing values
        :param dropnan: do not count nan values
        :param dropna: short for any of the above, (see :func:`Expression.isna`)
        :param bool axis: Axis over which to determine the unique elements (None will flatten arrays or lists)
        :param bool array_type: {array_type}
        """
        return self.ds.unique(self.expression, dropna=dropna, dropnan=dropnan, dropmissing=dropmissing, selection=selection, array_type=array_type, axis=axis, delay=delay)

    def nunique(self, dropna=False, dropnan=False, dropmissing=False, selection=None, axis=None, delay=False):
        """Counts number of unique values, i.e. `len(df.x.unique()) == df.x.nunique()`.

        :param dropmissing: do not count missing values
        :param dropnan: do not count nan values
        :param dropna: short for any of the above, (see :func:`Expression.isna`)
        :param bool axis: Axis over which to determine the unique elements (None will flatten arrays or lists)
        """
        return len(self.unique(dropna=dropna, dropnan=dropnan, dropmissing=dropmissing, selection=selection, axis=axis, delay=delay))

    def countna(self):
        """Returns the number of Not Availiable (N/A) values in the expression.
        This includes missing values and np.nan values.
        """
        return self.isna().sum().item()  # so the output is int, not array

    def countnan(self):
        """Returns the number of NaN values in the expression."""
        return self.isnan().sum().item()  # so the output is int, not array

    def countmissing(self):
        """Returns the number of missing values in the expression."""
        return self.ismissing().sum().item()  # so the output is int, not array

    def evaluate(self, i1=None, i2=None, out=None, selection=None, parallel=True, array_type=None):
        return self.ds.evaluate(self, i1, i2, out=out, selection=selection, array_type=array_type, parallel=parallel)

    # TODO: it is not so elegant we need to have a custom version of this
    # it now also misses the docstring, reconsider how the the meta class auto
    # adds this method
    def fillna(self, value, fill_nan=True, fill_masked=True):
        return self.ds.func.fillna(self, value=value, fill_nan=fill_nan, fill_masked=fill_masked)

    def clip(self, lower=None, upper=None):
        return self.ds.func.clip(self, lower, upper)

    def jit_numba(self, verbose=False):
        f = FunctionSerializableNumba.build(self.expression, df=self.ds, verbose=verbose, compile=self.ds.is_local())
        function = self.ds.add_function('_jit', f, unique=True)
        return function(*f.arguments)

    def jit_cuda(self, verbose=False):
        f = FunctionSerializableCuda.build(self.expression, df=self.ds, verbose=verbose, compile=self.ds.is_local())
        function = self.ds.add_function('_jit', f, unique=True)
        return function(*f.arguments)

    def jit_pythran(self, verbose=False):
        import logging
        logger = logging.getLogger('pythran')
        log_level = logger.getEffectiveLevel()
        try:
            if not verbose:
                logger.setLevel(logging.ERROR)
            import pythran
            import imp
            import hashlib
            # self._import_all(module)
            names = []
            funcs = set(expression_namespace.keys())
            expression = self.expression
            if expression in self.ds.virtual_columns:
                expression = self.ds.virtual_columns[self.expression]
            all_vars = self.ds.get_column_names(virtual=True, strings=True, hidden=True) + list(self.ds.variables.keys())
            vaex.expresso.validate_expression(expression, all_vars, funcs, names)
            names = list(set(names))
            types = ", ".join(str(self.ds.data_type(name)) + "[]" for name in names)
            argstring = ", ".join(names)
            code = '''
from numpy import *
#pythran export f({2})
def f({0}):
    return {1}'''.format(argstring, expression, types)
            if verbose:
                print("generated code")
                print(code)
            m = hashlib.md5()
            m.update(code.encode('utf-8'))
            module_name = "pythranized_" + m.hexdigest()
            # print(m.hexdigest())
            module_path = pythran.compile_pythrancode(module_name, code, extra_compile_args=["-DBOOST_SIMD", "-march=native"] + [] if verbose else ["-w"])

            module = imp.load_dynamic(module_name, module_path)
            function_name = "f_" + m.hexdigest()
            expression_namespace[function_name] = module.f

            return Expression(self.ds, "{0}({1})".format(function_name, argstring))
        finally:
                logger.setLevel(log_level)

    def _rename(self, old, new, inplace=False):
        expression = self if inplace else self.copy()
        if old in expression.ast_names:
            for node in expression.ast_names[old]:
                node.id = new
            expression._ast_names[new] = expression._ast_names.pop(old)
        slices = expression._ast_slices
        if old in slices:
            for node in slices[old]:
                node.slice.value.s = new
        expression._expression = None  # resets the cached string representation
        return expression

    def astype(self, data_type):
        if vaex.array_types.is_string_type(data_type) or data_type == str:
            return self.ds.func.astype(self, 'str')
        else:
            return self.ds.func.astype(self, str(data_type))

    def isin(self, values, use_hashmap=True):
        """Lazily tests if each value in the expression is present in values.

        :param values: List/array of values to check
        :param use_hashmap: use a hashmap or not (especially faster when values contains many elements)
        :return: :class:`Expression` with the lazy expression.
        """
        if use_hashmap:
            # easiest way to create a set is using the vaex dataframe
            df_values = vaex.from_arrays(x=values)
            ordered_set = df_values._set(df_values.x)
            var = self.df.add_variable('var_isin_ordered_set', ordered_set, unique=True)
            return self.df['isin_set(%s, %s)' % (self, var)]
        else:
            if self.is_string():
                values = vaex.column._to_string_sequence(values)
            else:
                values = np.array(values, dtype=self.dtype)
            var = self.df.add_variable('isin_values', values, unique=True)
            return self.df['isin(%s, %s)' % (self, var)]

    def apply(self, f, vectorize=False, multiprocessing=True):
        """Apply a function along all values of an Expression.

        Shorthand for ``df.apply(f, arguments=[expression])``, see :meth:`DataFrame.apply`

        Example:

        >>> df = vaex.example()
        >>> df.x
        Expression = x
        Length: 330,000 dtype: float64 (column)
        ---------------------------------------
             0  -0.777471
             1    3.77427
             2    1.37576
             3   -7.06738
             4   0.243441

        >>> def func(x):
        ...     return x**2

        >>> df.x.apply(func)
        Expression = lambda_function(x)
        Length: 330,000 dtype: float64 (expression)
        -------------------------------------------
             0   0.604461
             1    14.2451
             2    1.89272
             3    49.9478
             4  0.0592637


        :param f: A function to be applied on the Expression values
        :param vectorize: Call f with arrays instead of a scalars (for better performance).
        :param bool multiprocessing: Use multiple processes to avoid the GIL (Global interpreter lock).
        :returns: A function that is lazily evaluated when called.
        """
        return self.ds.apply(f, [self.expression], vectorize=vectorize, multiprocessing=multiprocessing)

    def dropmissing(self):
        # TODO: df.dropna does not support inplace
        # df = self.df if inplace else self.df.copy()
        df = self.ds
        df = df.dropmissing(column_names=[self.expression])
        return df._expr(self.expression)

    def dropnan(self):
        # TODO: df.dropna does not support inplace
        # df = self.df if inplace else self.df.copy()
        df = self.ds
        df = df.dropnan(column_names=[self.expression])
        return df._expr(self.expression)

    def dropna(self):
        # TODO: df.dropna does not support inplace
        # df = self.df if inplace else self.df.copy()
        df = self.ds
        df = df.dropna(column_names=[self.expression])
        return df._expr(self.expression)

    def map(self, mapper, nan_value=None, missing_value=None, default_value=None, allow_missing=False, axis=None):
        """Map values of an expression or in memory column according to an input
        dictionary or a custom callable function.

        Example:

        >>> import vaex
        >>> df = vaex.from_arrays(color=['red', 'red', 'blue', 'red', 'green'])
        >>> mapper = {'red': 1, 'blue': 2, 'green': 3}
        >>> df['color_mapped'] = df.color.map(mapper)
        >>> df
        #  color      color_mapped
        0  red                   1
        1  red                   1
        2  blue                  2
        3  red                   1
        4  green                 3
        >>> import numpy as np
        >>> df = vaex.from_arrays(type=[0, 1, 2, 2, 2, np.nan])
        >>> df['role'] = df['type'].map({0: 'admin', 1: 'maintainer', 2: 'user', np.nan: 'unknown'})
        >>> df
        #    type  role
        0       0  admin
        1       1  maintainer
        2       2  user
        3       2  user
        4       2  user
        5     nan  unknown
        >>> import vaex
        >>> import numpy as np
        >>> df = vaex.from_arrays(type=[0, 1, 2, 2, 2, 4])
        >>> df['role'] = df['type'].map({0: 'admin', 1: 'maintainer', 2: 'user'}, default_value='unknown')
        >>> df
        #    type  role
        0       0  admin
        1       1  maintainer
        2       2  user
        3       2  user
        4       2  user
        5       4  unknown
        :param mapper: dict like object used to map the values from keys to values
        :param nan_value: value to be used when a nan is present (and not in the mapper)
        :param missing_value: value to use used when there is a missing value
        :param default_value: value to be used when a value is not in the mapper (like dict.get(key, default))
        :param allow_missing: used to signal that values in the mapper should map to a masked array with missing values,
            assumed True when default_value is not None.
        :param bool axis: Axis over which to determine the unique elements (None will flatten arrays or lists)
        :return: A vaex expression
        :rtype: vaex.expression.Expression
        """
        assert isinstance(mapper, collectionsAbc.Mapping), "mapper should be a dict like object"
        if axis is not None:
            raise ValueError('only axis=None is supported')

        df = self.ds
        mapper_keys = list(mapper.keys())
        try:
            mapper_nan_key_mask = np.isnan(mapper_keys)
        except TypeError:
            # case where we have mixed strings/nan etc
            def try_nan(x):
                try:
                    return np.isnan(x)
                except:
                    return False
            mapper_nan_key_mask = np.array([try_nan(k) for k in mapper_keys])
        mapper_has_nan = mapper_nan_key_mask.sum() > 0
        if mapper_nan_key_mask.sum() > 1:
            raise ValueError('Insanity, you provided multiple nan values as keys for your dict')
        if mapper_has_nan:
            for key, value in mapper.items():
                if key != key:
                    nan_value = value
        for key, value in mapper.items():
            if key is None:
                missing_value = value

        if axis is not None:
            raise ValueError('only axis=None is supported')
        # we map the keys to a ordinal values [0, N-1] using the set
        key_set = df._set(self.expression, flatten=axis is None)
        found_keys = key_set.keys()

        # we want all possible values to be converted
        # so mapper's key should be a superset of the keys found
        use_masked_array = False
        if default_value is not None:
            allow_missing = True
        if allow_missing:
            use_masked_array = True
        if not set(mapper_keys).issuperset(found_keys):
            missing = set(found_keys).difference(mapper_keys)
            missing0 = list(missing)[0]
            only_has_nan = missing0 != missing0 and len(missing) == 1
            if allow_missing:
                if default_value is not None:
                    value0 = list(mapper.values())[0]
                    assert np.issubdtype(type(default_value), np.array(value0).dtype), "default value has to be of similar type"
            else:
                if only_has_nan:
                    pass  # we're good, the hash mapper deals with nan
                else:
                    raise ValueError('Missing %i values in mapper: %s' % (len(missing), missing))

        # and these are the corresponding choices
        # note that here we map 'planned' unknown values to the default values
        # and later on in _choose, we map values not even seen in the dataframe
        # to the default_value
        dtype = self.data_type(self.expression)
        dtype_item = self.data_type(self.expression, axis=-1)
        if dtype_item.is_float:
            print(nan_value)
            values  = [np.nan, None] + [key for key in mapper if key == key and key is not None]
            choices = [default_value, nan_value, missing_value] + [mapper[key] for key in mapper if key == key and key is not None]
        else:
            values  = [None] + [key for key in mapper if key is not None]
            choices = [default_value, missing_value] + [mapper[key] for key in mapper if key is not None]
        values = pa.array(values)
        choices = pa.array(choices)
        from .hash import ordered_set_type_from_dtype
        ordered_set_type = ordered_set_type_from_dtype(dtype_item)
        ordered_set = ordered_set_type()
        if vaex.array_types.is_string_type(dtype_item):
            values = _to_string_sequence(values)
        else:
            values = vaex.array_types.to_numpy(values)
        if np.ma.isMaskedArray(values):
            mask = np.ma.getmaskarray(values)
            ordered_set.update(values.data, mask)
        else:
            ordered_set.update(values)

        key_set_name = df.add_variable('map_key_set', ordered_set, unique=True)
        choices_name = df.add_variable('map_choices', choices, unique=True)
        if allow_missing:
            expr = '_map({}, {}, {}, use_missing={!r}, axis={!r})'.format(self, key_set_name, choices_name, use_masked_array, axis)
        else:
            expr = '_map({}, {}, {}, axis={!r})'.format(self, key_set_name, choices_name, axis)
        return Expression(df, expr)

    @property
    def is_masked(self):
        return self.ds.is_masked(self.expression)

    def is_string(self):
        return self.df.is_string(self.expression)


class FunctionSerializable(object):
    pass


@vaex.serialize.register
class FunctionSerializablePickle(FunctionSerializable):
    def __init__(self, f=None, multiprocessing=False):
        self.f = f
        self.multiprocessing = multiprocessing

    def pickle(self, function):
        return pickle.dumps(function)

    def unpickle(self, data):
        return pickle.loads(data)

    def __getstate__(self):
        return self.state_get()

    def __setstate__(self, state):
        self.state_set(state)

    def state_get(self):
        data = self.pickle(self.f)
        if vaex.utils.PY2:
            pickled = base64.encodestring(data)
        else:
            pickled = base64.encodebytes(data).decode('ascii')
        return dict(pickled=pickled)

    @classmethod
    def state_from(cls, state, trusted=True):
        obj = cls()
        obj.state_set(state, trusted=trusted)
        return obj

    def state_set(self, state, trusted=True):
        data = state['pickled']
        if vaex.utils.PY2:
            data = base64.decodestring(data)
        else:
            data = base64.decodebytes(data.encode('ascii'))
        if trusted is False:
            raise ValueError("Will not unpickle data when source is not trusted")
        self.f = self.unpickle(data)

    def __call__(self, *args, **kwargs):
        '''Forward the call to the real function'''
        import vaex.multiprocessing
        return vaex.multiprocessing.apply(self._apply, args, kwargs, self.multiprocessing)

    def _apply(self, *args, **kwargs):
        return self.f(*args, **kwargs)


class FunctionSerializableJit(FunctionSerializable):
    def __init__(self, expression, arguments, argument_dtypes, return_dtype, verbose=False, compile=True):
        self.expression = expression
        self.arguments = arguments
        self.argument_dtypes = argument_dtypes
        self.return_dtype = return_dtype
        self.verbose = verbose
        if compile:
            self.f = self.compile()
        else:
            def placeholder(*args, **kwargs):
                raise Exception('You chose not to compile this function (locally), but did invoke it')
            self.f = placeholder

    def state_get(self):
        return dict(expression=self.expression,
                    arguments=self.arguments,
                    argument_dtypes=list(map(lambda dtype: str(dtype.numpy), self.argument_dtypes)),
                    return_dtype=str(self.return_dtype),
                    verbose=self.verbose)

    @classmethod
    def state_from(cls, state, trusted=True):
        return cls(expression=state['expression'],
                   arguments=state['arguments'],
                   argument_dtypes=list(map(lambda s: DataType(np.dtype(s)), state['argument_dtypes'])),
                   return_dtype=DataType(np.dtype(state['return_dtype'])),
                   verbose=state['verbose'])

    @classmethod
    def build(cls, expression, df=None, verbose=False, compile=True):
        df = df or expression.df
        # if it's a virtual column, we probably want to optimize that
        # TODO: fully extract the virtual columns, i.e. depending ones?
        expression = str(expression)

        if expression in df.virtual_columns:
            expression = df.virtual_columns[expression]

        # function validation, and finding variable names
        all_vars = df.get_column_names(hidden=True) + list(df.variables.keys())
        funcs = set(list(expression_namespace.keys()) + list(df.functions.keys()))
        names = []
        vaex.expresso.validate_expression(expression, all_vars, funcs, names)
        # TODO: can we do the above using the Expressio API?s

        arguments = list(set(names))
        argument_dtypes = [df.data_type(argument, array_type='numpy') for argument in arguments]
        return_dtype = df[expression].dtype
        return cls(str(expression), arguments, argument_dtypes, return_dtype, verbose, compile=compile)

    def __call__(self, *args, **kwargs):
        '''Forward the call to the numba function'''
        return self.f(*args, **kwargs)


@vaex.serialize.register
class FunctionSerializableNumba(FunctionSerializableJit):
    def compile(self):
        import numba
        argstring = ", ".join(self.arguments)
        code = '''
from numpy import *
def f({0}):
    return {1}'''.format(argstring, self.expression)
        if self.verbose:
            print('Generated code:\n' + code)
        scope = {}
        exec(code, scope)
        f = scope['f']

        # numba part
        argument_dtypes_numba = [getattr(numba, argument_dtype.numpy.name) for argument_dtype in self.argument_dtypes]
        return_dtype_numba = getattr(numba, self.return_dtype.numpy.name)
        vectorizer = numba.vectorize([return_dtype_numba(*argument_dtypes_numba)])
        return vectorizer(f)


@vaex.serialize.register
class FunctionSerializableCuda(FunctionSerializableJit):
    def compile(self):
        import cupy
        # code generation
        argstring = ", ".join(self.arguments)
        code = '''
from cupy import *
import cupy
@fuse()
def f({0}):
    return {1}
'''.format(argstring, self.expression)#, ";".join(conversions))
        if self.verbose:
            print("generated code")
            print(code)
        scope = dict()#cupy=cupy)
        exec(code, scope)
        func = scope['f']
        def wrapper(*args):
            args = [vaex.array_types.to_numpy(k) for k in args]
            args = [vaex.utils.to_native_array(arg) if isinstance(arg, np.ndarray) else arg for arg in args]
            args = [cupy.asarray(arg) if isinstance(arg, np.ndarray) else arg for arg in args]
            return cupy.asnumpy(func(*args))
        return wrapper


# TODO: this is not the right abstraction, since this won't allow a
# numba version for the function
@vaex.serialize.register
class FunctionToScalar(FunctionSerializablePickle):
    def __call__(self, *args, **kwargs):
        import vaex.multiprocessing
        return vaex.multiprocessing.apply(self._apply, args, kwargs, self.multiprocessing)

    def _apply(self, *args, **kwargs):
        length = len(args[0])
        result = []
        def fix_type(v):
            # TODO: only when column is str type?
            if isinstance(v, np.str_):
                return str(v)
            if isinstance(v, np.bytes_):
                return v.decode('utf8')
            else:
                return v
        args = [vaex.array_types.tolist(k) for k in args]
        for i in range(length):
            scalar_result = self.f(*[fix_type(k[i]) for k in args], **{key: value[i] for key, value in kwargs.items()})
            result.append(scalar_result)
        result = np.array(result)
        return result


class Function(object):

    def __init__(self, dataset, name, f):
        self.dataset = dataset
        self.name = name

        if not vaex.serialize.can_serialize(f): # if not serializable, assume we can use pickle
            f = FunctionSerializablePickle(f)
        self.f = f

    def __call__(self, *args, **kwargs):
        arg_string = ", ".join([str(k) for k in args] + ['{}={:r}'.format(name, value) for name, value in kwargs.items()])
        expression = "{}({})".format(self.name, arg_string)
        return Expression(self.dataset, expression)


class FunctionBuiltin(object):

    def __init__(self, dataset, name, **kwargs):
        self.dataset = dataset
        self.name = name
        self.kwargs = kwargs

    def __call__(self, *args, **kwargs):
        kwargs = dict(kwargs, **self.kwargs)
        arg_string = ", ".join([str(k) for k in args] + ['{}={:r}'.format(name, value) for name, value in kwargs.items()])
        expression = "{}({})".format(self.name, arg_string)
        return Expression(self.dataset, expression)



import re
import logging
import os

import vaex
import vaex.file
import vaex.cache

log = logging.getLogger('vaex.cache')


def _convert_name(filenames, shuffle=False, suffix=None):
    '''Convert a filename (or list of) to a filename with .hdf5 and optionally a -shuffle or other suffix'''
    if not isinstance(filenames, (list, tuple)):
        filenames = [filenames]
    filenames = [vaex.file.stringyfy(f) for f in filenames]
    base = filenames[0]
    if shuffle:
        base += '-shuffle'
    if suffix:
        base += suffix
    if len(filenames) > 1:
        return base + "_and_{}_more.hdf5".format(len(filenames)-1)
    else:
        return base + ".hdf5"


def convert(path_input, fs_options_input, fs_input, path_output, fs_options_output, fs_output, *args, **kwargs):
    @vaex.cache.output_file(
        path_input=path_input, fs_options_input=fs_options_input, fs_input=fs_input,
        path_output=path_output, fs_options_output=fs_options_output, fs_output=fs_output)
    def cached_output(*args, **kwargs):
        ds = vaex.dataset.open(path_input, fs_options=fs_options_input, fs=fs_input, *args, **kwargs)
        if ds is not None:
            df = vaex.from_dataset(ds)
            df.export(path_output, fs_options=fs_options_output, fs=fs_output)
    cached_output(*args, **kwargs)


def convert_csv(path_input, fs_options_input, fs_input, path_output, fs_options_output, fs_output, *args, **kwargs):
    @vaex.cache.output_file(
        path_input=path_input, fs_options_input=fs_options_input, fs_input=fs_input,
        path_output=path_output, fs_options_output=fs_options_output, fs_output=fs_output)
    def cached_output(*args, **kwargs):
        if fs_options_output:
            raise ValueError(f'No fs_options support for output/convert')
        if 'chunk_size' not in kwargs:
            # make it memory efficient by default
            kwargs['chunk_size'] = 5_000_000
        _from_csv_convert_and_read(path_input, path_output=path_output, fs_options=fs_options_input, fs=fs_input, **kwargs)
    cached_output(*args, **kwargs)


def _from_csv_convert_and_read(filename_or_buffer, path_output, chunk_size, fs_options, fs=None, copy_index=False, **kwargs):
    # figure out the CSV file path
    csv_path = vaex.file.stringyfy(filename_or_buffer)
    path_output_bare, ext, _ = vaex.file.split_ext(path_output)

    combined_hdf5 = _convert_name(csv_path)

    # convert CSV chunks to separate HDF5 files
    import pandas as pd
    converted_paths = []
    with vaex.file.open(filename_or_buffer, fs_options=fs_options, fs=fs, for_arrow=True) as f:
        csv_reader = pd.read_csv(filename_or_buffer, chunksize=chunk_size, **kwargs)
        for i, df_pandas in enumerate(csv_reader):
            df = vaex.from_pandas(df_pandas, copy_index=copy_index)
            chunk_name = f'{path_output_bare}_chunk_{i}.{ext}'
            df.export(chunk_name)
            converted_paths.append(chunk_name)
            log.info('saved chunk #%d to %s' % (i, chunk_name))

    # combine chunks into one HDF5 file
    if len(converted_paths) == 1:
        # no need to merge several HDF5 files
        os.rename(converted_paths[0], path_output)
    else:
        log.info('converting %d chunks into single file %s' % (len(converted_paths), path_output))
        dfs = [vaex.open(p) for p in converted_paths]
        df_combined = vaex.concat(dfs)
        df_combined.export(path_output)

        log.info('deleting %d chunk files' % len(converted_paths))
        for df, df_path in zip(dfs, converted_paths):
            try:
                df.close()
                os.remove(df_path)
            except Exception as e:
                log.error('Could not close or delete intermediate file %s used to convert %s to single file: %s', (df_path, csv_path, path_output))


def main(argv):
    import argparse
    parser = argparse.ArgumentParser(argv[0])
    parser.add_argument('--verbose', '-v', action='count', default=0)
    parser.add_argument('--quiet', '-q', default=False, action='store_true', help="do not output anything")
    parser.add_argument('--list', '-l', default=False, action='store_true', help="list columns of input")
    parser.add_argument('--progress', help="show progress (default: %(default)s)", default=True, action='store_true')
    parser.add_argument('--no-progress', dest="progress", action='store_false')
    parser.add_argument('--no-delete', help="Delete file on failure (default: %(default)s)", dest='delete', default=True, action='store_false')
    parser.add_argument('--shuffle', "-s", dest="shuffle", action='store_true', default=False)
    parser.add_argument('--sort', dest="sort", default=None)
    parser.add_argument('--fraction', "-f", dest="fraction", type=float, default=1.0, help="fraction of input dataset to export")
    parser.add_argument('--filter', dest="filter", default=None, help="filter to apply before exporting")
    parser.add_argument("input", help="input source or file, when prefixed with @ it is assumed to be a text file with a file list (one file per line)")
    parser.add_argument("output", help="output file (ends in .hdf5)")
    parser.add_argument("columns", help="list of columns to export (or all when empty)", nargs="*")

    args = parser.parse_args(argv[1:])

    verbosity = ["ERROR", "WARNING", "INFO", "DEBUG"]
    logging.getLogger("vaex").setLevel(verbosity[min(3, args.verbose)])
    if args.input[0] == "@":
        inputs = open(args.input[1:]).readlines()
        df = vaex.open_many(inputs)
    else:
        df = vaex.open(args.input)

    if df:
        df.set_active_fraction(args.fraction)
    if args.list:
        print("\n".join(df.get_column_names()))
    else:
        if args.columns:
            all_columns = df.get_column_names()
            columns = args.columns
            for column in columns:
                if column not in all_columns:
                    # if not args.quiet:
                    print("column %r does not exist, run with --list or -l to list all columns" % column)
                    return 1
            df = df[columns]
        else:
            columns = df.get_column_names()

        if not args.quiet:
            print("exporting %d rows and %d columns" % (len(df), len(columns)))
            print("columns: " + " ".join(columns))

        if args.filter:
            df = df.filter(args.filter)
        if args.sort:
            df = df.sort(args.sort)
        try:
            df.export(args.output, progress=args.progress)
            if not args.quiet:
                print("\noutput to %s" % os.path.abspath(args.output))
            df.close()
        except:
            if not args.quiet:
                print("\nfailed to write to%s" % os.path.abspath(args.output))
            if args.delete:
                if args.delete:
                    os.remove(args.output)
                    print("\ndeleted output %s (pass --no-delete to avoid that)" % os.path.abspath(args.output))
            raise


    return 0



import logging
import numpy as np

import vaex.promise
import vaex.encoding
from .utils import _expand_shape
from .datatype import DataType


logger = logging.getLogger('vaex.tasks')
_task_register = {}


@vaex.encoding.register("task")
class task_encoder:
    @staticmethod
    def encode(encoding, task):
        return task.encode(encoding)

    @staticmethod
    def decode(encoding, spec, df):
        cls = _task_register[spec['task']]
        return cls.decode(encoding, spec, df)


def register(cls):
    _task_register[cls.name] = cls
    return cls


class Task(vaex.promise.Promise):
    """
    :type: signal_progress: Signal
    """

    def __init__(self, df=None, expressions=[], pre_filter=False, name="task"):
        vaex.promise.Promise.__init__(self)
        self.df = df
        self.expressions = expressions
        self.expressions_all = list(expressions)
        self.signal_progress = vaex.events.Signal("progress (float)")
        self.progress_fraction = 0
        self.signal_progress.connect(self._set_progress)
        self.cancelled = False
        self.name = name
        self.pre_filter = pre_filter
        self.result = None

    def _set_progress(self, fraction):
        self.progress_fraction = fraction
        return not self.cancelled  # don't cancel

    def cancel(self):
        self.cancelled = True

    @property
    def dimension(self):
        return len(self.expressions)

    @classmethod
    def create(cls):
        ret = Task()
        return ret

    def create_next(self):
        ret = Task(self.df, [])
        self.signal_progress.connect(ret.signal_progress.emit)
        return ret

    # def __repr__(self):
    #     from .encoding import Encoding
    #     encoding = Encoding()
    #     return json.dumps({key: repr(value) for key, value in self.encode(encoding).items()})


# only used for testing
@register
class TaskSum(Task):
    name = "sum-test"

    # def __init__(self, df, expression):
    #     super().__init__(df, expression)
    #     self.expression = expression

    def encode(self, encoding):
        return {'task': type(self).name, 'expression': self.expressions}

    @classmethod
    def decode(cls, encoding, spec, df):
        return cls(df, expression=[spec['expression']])


@register
class TaskMapReduce(Task):
    name = "map_reduce"

    def __init__(self, df, expressions, map, reduce, info=False, to_float=False,
                 to_numpy=True, skip_masked=False, ignore_filter=False, selection=None, pre_filter=False, name="task"):
        Task.__init__(self, df, expressions, name=name, pre_filter=pre_filter)
        self._map = map
        self._reduce = reduce
        self.info = info
        self.to_float = to_float
        self.to_numpy = to_numpy
        self.skip_masked = skip_masked
        self.ignore_filter = ignore_filter
        if self.pre_filter and self.ignore_filter:
            raise ValueError("Cannot pre filter and also ignore the filter")
        self.selection = selection

    def encode(self, encoding):
        return {'task': type(self).name, 'expressions': self.expressions, 'map': self._map, 'reduce': self._reduce,
                'info': self.info, 'to_float': self.to_float, 'to_numpy': self.to_numpy,  # 'ordered_reduce': self.ordered_reduce,
                'skip_masked': self.skip_masked, 'ignore_filter': self.ignore_filter, 'selection': self.selection, 'pre_filter': self.pre_filter,
                }


class StatOp(object):
    def __init__(self, code, fields, reduce_function=np.nansum, dtype=None):
        self.code = code
        self.fixed_fields = fields
        self.reduce_function = reduce_function
        self.dtype = dtype

    def init(self, grid):
        pass

    def fields(self, weights):
        return self.fixed_fields

    def reduce(self, grid, axis=0):
        value = self.reduce_function(grid, axis=axis)
        if self.dtype:
            return value.astype(self.dtype)
        else:
            return value

    def encode(self, encoding):
        return {'code': self.code, 'fields': self.fixed_fields, 'reduce_function': self.reduce_function.__name__}


class StatOpMinMax(StatOp):
    def __init__(self, code, fields):
        super(StatOpMinMax, self).__init__(code, fields)

    def init(self, grid):
        grid[..., 0] = np.inf
        grid[..., 1] = -np.inf

    def reduce(self, grid, axis=0):
        out = np.zeros(grid.shape[1:], dtype=grid.dtype)
        out[..., 0] = np.nanmin(grid[..., 0], axis=axis)
        out[..., 1] = np.nanmax(grid[..., 1], axis=axis)
        return out

    def encode(self, encoding):
        return {'code': self.code, 'fields': self.fixed_fields}


class StatOpCov(StatOp):
    def __init__(self, code, fields=None, reduce_function=np.sum):
        super(StatOpCov, self).__init__(code, fields, reduce_function=reduce_function)

    def fields(self, weights):
        N = len(weights)
        # counts, sums, cross product sums
        return N * 2 + N**2 * 2  # ((N+1) * N) // 2 *2

    def encode(self, encoding):
        return {'code': self.code, 'reduce_function': self.reduce_function.__name__}


class StatOpFirst(StatOp):
    def __init__(self, code):
        super(StatOpFirst, self).__init__(code, 2, reduce_function=self._reduce_function)

    def init(self, grid):
        grid[..., 0] = np.nan
        grid[..., 1] = np.inf

    def _reduce_function(self, grid, axis=0):
        values = grid[..., 0]
        order_values = grid[..., 1]
        indices = np.argmin(order_values, axis=0)

        # see e.g. https://stackoverflow.com/questions/46840848/numpy-how-to-use-argmax-results-to-get-the-actual-max?noredirect=1&lq=1
        # and https://jakevdp.github.io/PythonDataScienceHandbook/02.07-fancy-indexing.html
        if len(values.shape) == 2:  # no binby
            return values[indices, np.arange(values.shape[1])[:, None]][0]
        if len(values.shape) == 3:  # 1d binby
            return values[indices, np.arange(values.shape[1])[:, None], np.arange(values.shape[2])]
        if len(values.shape) == 4:  # 2d binby
            return values[indices, np.arange(values.shape[1])[:, None], np.arange(values.shape[2])[None, :, None], np.arange(values.shape[3])]
        else:
            raise ValueError('dimension %d not yet supported' % len(values.shape))

    def fields(self, weights):
        # the value found, and the value by which it is ordered
        return 2

    def encode(self, encoding):
        return {'code': self.code, 'reduce_function': self.reduce_function.__name__}


@vaex.encoding.register('_op')
class op_encoding:
    @staticmethod
    def encode(encoding, op):
        return op.encode(encoding)

    @staticmethod
    def decode(encoding, op_spec):
        op_spec = op_spec.copy()
        if 'reduce_function' in op_spec:
            op_spec['reduce_function'] = getattr(np, op_spec.pop('reduce_function'))
        cls = StatOp
        if op_spec['code'] == 2:
            cls = StatOpMinMax
        if op_spec['code'] == 5:
            cls = StatOpCov
        if op_spec['code'] == 6:
            cls = StatOpFirst
        return cls(**op_spec)


OP_ADD1 = StatOp(0, 1)
OP_COUNT = StatOp(1, 1)
OP_MIN_MAX = StatOpMinMax(2, 2)
OP_ADD_WEIGHT_MOMENTS_01 = StatOp(3, 2, np.nansum)
OP_ADD_WEIGHT_MOMENTS_012 = StatOp(4, 3, np.nansum)
OP_COV = StatOpCov(5)
OP_FIRST = StatOpFirst(6)


@register
class TaskStatistic(Task):
    name = "legacy_statistic"

    def encode(self, encoding):
        return {'task': type(self).name, 'expressions': self.expressions,
                'shape': self.shape, 'selections': self.selections, 'op': encoding.encode('_op', self.op), 'weights': self.weights,
                'dtype': encoding.encode('dtype', DataType(self.dtype)), 'minima': self.minima, 'maxima': self.maxima, 'edges': self.edges,
                'selection_waslist': self.selection_waslist}

    @classmethod
    def decode(cls, encoding, spec, df):
        spec = spec.copy()
        del spec['task']
        spec['op'] = encoding.decode('_op', spec['op'])
        spec['dtype'] = encoding.decode('dtype', spec['dtype'])
        selection_waslist = spec.pop('selection_waslist')
        if selection_waslist:
            spec['selection'] = spec.pop('selections')
        else:
            spec['selection'] = spec.pop('selections')[0]
        spec['limits'] = list(zip(spec.pop('minima'), spec.pop('maxima')))
        return cls(df, **spec)

    def __init__(self, df, expressions, shape, limits, masked=False, weights=[], weight=None, op=OP_ADD1, selection=None, edges=False,
                 dtype=np.dtype('f8')):
        if not isinstance(expressions, (tuple, list)):
            expressions = [expressions]
        # edges include everything outside at index 1 and -1, and nan's at index 0, so we add 3 to each dimension
        self.shape = tuple([k + 3 if edges else k for k in _expand_shape(shape, len(expressions))])
        self.limits = limits
        if weight is not None:  # shortcut for weights=[weight]
            assert weights == [], 'only provide weight or weights, not both'
            weights = [weight]
            del weight
        self.weights = weights
        self.selection_waslist, [self.selections, ] = vaex.utils.listify(selection)
        self.op = op
        self.edges = edges
        Task.__init__(self, df, expressions, name="statisticNd", pre_filter=df.filtered)
        # self.dtype = np.int64 if self.op == OP_ADD1 else np.float64 # TODO: use int64 fir count and ADD1
        self.dtype = dtype
        self.masked = masked

        self.fields = op.fields(weights)
        # self.shape_total = (self.df.executor.thread_pool.nthreads,) + (len(self.selections), ) + self.shape + (self.fields,)
        # self.grid = np.zeros(self.shape_total, dtype=self.dtype)
        # self.op.init(self.grid)
        self.minima = []
        self.maxima = []
        limits = np.array(self.limits)
        if len(limits) != 0:
            logger.debug("limits = %r", limits)
            assert limits.shape[-1] == 2, "expected last dimension of limits to have a length of 2 (not %d, total shape: %s), of the form [[xmin, xmin], ... [zmin, zmax]], not %s" %\
                                          (limits.shape[-1], limits.shape, limits)
            if len(limits.shape) == 1:  # short notation: [xmin, max], instead of [[xmin, xmax]]
                limits = [limits]
            logger.debug("limits = %r", limits)
            for limit in limits:
                vmin, vmax = limit
                self.minima.append(float(vmin))
                self.maxima.append(float(vmax))
        # if self.weight is not None:
        self.expressions_all.extend(weights)


@register
class TaskAggregations(Task):
    """Multiple aggregations on a single grid."""
    name = "aggregations"

    def __init__(self, df, grid):
        expressions = [binner.expression for binner in grid.binners]
        self.df = df
        self.parent_grid = grid
        self.aggregation_descriptions = []
        self.dtypes = {}
        Task.__init__(self, df, expressions, name="statisticNd", pre_filter=df.filtered)

    def encode(self, encoding):
        # TODO: get rid of dtypes
        return {'task': type(self).name,
                'grid': encoding.encode('grid', self.parent_grid),
                'aggregations': encoding.encode_list("aggregation", self.aggregation_descriptions),
                'dtypes': encoding.encode_dict("dtype", self.dtypes)
                }

    @classmethod
    def decode(cls, encoding, spec, df):
        grid = encoding.decode('grid', spec['grid'])
        task = cls(df, grid)
        aggs = encoding.decode_list('aggregation', spec['aggregations'])
        for agg in aggs:
            agg._prepare_types(df)
            task.add_aggregation_operation(agg)
        return task

    def add_aggregation_operation(self, aggregator_descriptor):
        task = Task(self.df, [], "--")

        def chain_reject(x):
            task.reject(x)
            return x

        def assign_subtask(values, index=len(self.aggregation_descriptions)):
            task.fulfill(values[index])
        self.then(assign_subtask, chain_reject)

        self.aggregation_descriptions.append((aggregator_descriptor))
        # THIS SHOULD BE IN THE SAME ORDER AS THE ABOVE TASKPART
        # it is up the the executor to remove duplicate expressions
        self.expressions_all.extend(aggregator_descriptor.expressions)
        # TODO: optimize/remove?
        self.dtypes = {expr: self.df.data_type(expr) for expr in self.expressions_all}
        return task

    def check(self):
        if not self.aggregation_descriptions:
            raise RuntimeError('Aggregation tasks started but nothing to do, maybe adding operations failed?')



# -*- coding: utf-8 -*-
import sys
import os
import platform

darwin = "darwin" in platform.system().lower()
frozen = getattr(sys, 'frozen', False)
if frozen and len(sys.argv) > 1 and sys.argv[1].startswith("-psn"):  # is the 'app' is opened in osx, just stars without arguments
    dirname = os.path.join(os.path.dirname(sys.argv[0]), "..", "..", "..")
    os.chdir(dirname)


if darwin and frozen:  # on newer osx versions we get a lot of broken pipes when writing to stdout
    directory = os.path.expanduser("~/.vaex")
    if not os.path.exists(directory):
        os.makedirs(directory)
    sys.stdout = open(os.path.expanduser('~/.vaex/stdout.txt'), 'w')
    sys.stderr = open(os.path.expanduser('~/.vaex/stderr.txt'), 'w')


# print darwin, platform.system()


usage = """usage veax [-h] {webserver,convert,...}

optional arguments:
  -h, --help          show this help message and exit

positional arguments:
    webserver           start the vaex webserver
    convert             convert datasets
    benchmark           run benchmarks
    meta                import/export meta data
    alias               manage aliases
    stat                print statistics/info about dataset
    test                run unittests
    open                tests opening of a file (will return exit error on failure)
    ...                 anything else will start up the gui, see usage below

Examples:
$ vaex convert s3://vaex/taxi/nyc_taxi_2015_mini.parquet taxi.hdf5
$ vaex convert taxi.hdf5 taxi.csv

"""


def main(args=None):
    if args is None:
        args = sys.argv
    if frozen and len(args) > 1 and args[1].startswith("-psn"):  # is the 'app' is opened in osx, just start without arguments
        import vaex.ui.main
        vaex.ui.main.main([])
    else:
        if len(args) > 1 and args[1] in ["-h", "--help"]:
            print(usage)
            sys.exit(0)
        if len(args) > 1 and args[1] == "version":
            import vaex.version
            if frozen:
                extra = " (build on %s using Python %s)" % (vaex.version.osname, sys.version)
            else:
                extra = " (using Python %s)" % (sys.version)
            print(vaex.__full_name__ + extra)
        elif len(args) > 1 and args[1] == "webserver":
            import vaex.server.tornado_server
            vaex.server.tornado_server.main([os.path.basename(args[0]) + " " + args[1]] + args[2:])
        elif len(args) > 1 and args[1] == "benchmark":
            import vaex.benchmark
            vaex.benchmark.main([os.path.basename(args[0]) + " " + args[1]] + args[2:])
        elif len(args) > 1 and args[1] == "convert":
            import vaex.convert
            vaex.convert.main([os.path.basename(args[0]) + " " + args[1]] + args[2:])
        elif len(args) > 1 and args[1] == "meta":
            import vaex.meta
            vaex.meta.main([os.path.basename(args[0]) + " " + args[1]] + args[2:])
        elif len(args) > 1 and args[1] == "alias":
            import vaex.misc_cmdline
            vaex.misc_cmdline.alias_main([os.path.basename(args[0]) + " " + args[1]] + args[2:])
        elif len(args) > 1 and args[1] == "stat":
            import vaex.misc_cmdline
            vaex.misc_cmdline.stat_main([os.path.basename(args[0]) + " " + args[1]] + args[2:])
        elif len(args) > 1 and args[1] == "open":
            sys.exit(open_main([os.path.basename(args[0]) + " " + args[1]] + args[2:]))
        elif len(args) > 1 and args[1] == "test":
            import vaex.test.__main__
            vaex.test.__main__.main([os.path.basename(args[0]) + " " + args[1]] + args[2:])
        else:
            print(usage)
            sys.exit(0)

def open_main(argv):
    import argparse
    parser = argparse.ArgumentParser(argv[0])
    parser.add_argument('--verbose', '-v', default=False, action='store_true', help="give extra output")
    parser.add_argument('--quiet', '-q', default=False, action='store_true', help="do not output anything")
    parser.add_argument('--dry-run', '-n', default=False, action='store_true', help="do not actually execute commands (like delete)")
    parser.add_argument('--delete', help="Delete file when reading fails", default=False, action='store_true')
    parser.add_argument("input", help="list of files to try to open", nargs="*")

    args = parser.parse_args(argv[1:])
    import vaex
    import vaex.file
    failed = False
    if args.verbose:
        print(f"Checking files {', '.join(args.input)}")
    for path in args.input:
        try:
            vaex.open(path)
        except BaseException as e:
            failed = True
            if not args.quiet:
                print(e)
            if args.delete:
                if not args.quiet:
                    print(f'rm {path}')
                if not args.dry_run:
                    try:
                        vaex.file.remove(path)
                    except FileNotFoundError:
                        pass
    if args.verbose:
        if failed:
            print("Oops, had issues opening some files")
        else:
            print("All files could be opened")
    return 123 if failed else 0

if __name__ == "__main__":
    main()



from abc import  abstractmethod
import os
from pathlib import Path
import collections.abc
import logging
import pkg_resources
import uuid
from urllib.parse import urlparse

import numpy as np
import blake3
from frozendict import frozendict
import pyarrow as pa

import vaex
from vaex.array_types import data_type
from .column import Column, ColumnIndexed, supported_column_types
from . import array_types

logger = logging.getLogger('vaex.dataset')

opener_classes = []
HASH_VERSION = "1"


def open(path, fs_options={}, fs=None, *args, **kwargs):
    failures = []
    if not opener_classes:
        for entry in pkg_resources.iter_entry_points(group='vaex.dataset.opener'):
            logger.debug('trying opener: ' + entry.name)
            try:
                opener = entry.load()
                opener_classes.append(opener)
            except Exception as e:
                logger.exception('issue loading ' + entry.name)
                failures.append((e, entry))

    # first the quick path
    for opener in opener_classes:
        if opener.quick_test(path, fs_options=fs_options, fs=fs):
            if opener.can_open(path, fs_options=fs_options, fs=fs, *args, **kwargs):
                return opener.open(path, fs_options=fs_options, fs=fs, *args, **kwargs)

    # otherwise try all openers
    for opener in opener_classes:
        try:
            if opener.can_open(path, fs_options=fs_options, fs=fs, *args, **kwargs):
                return opener.open(path, fs_options=fs_options, fs=fs, *args, **kwargs)
        except Exception as e:
            failures.append((e, opener))

    failures = "\n".join([f'\n-----{who}-----\n:' + vaex.utils.format_exception_trace(e) for e, who in failures])
    if failures:
        raise IOError(f'Cannot open {path}, failures: {failures}.')
    else:
        raise IOError(f'Cannot open {path} nobody knows how to read it.')


def _to_bytes(ar):
    try:
        return ar.view(np.uint8)
    except ValueError:
        return ar.copy().view(np.uint8)

def hash_combine(*hashes):
    blake = blake3.blake3(multithreading=False)
    for hash in hashes:
        blake.update(hash.encode())
    return blake.hexdigest()


def hash_slice(hash, start, end):
    blake = blake3.blake3(hash.encode(), multithreading=False)
    slice = np.array([start, end], dtype=np.int64)
    blake.update(_to_bytes(slice))
    return blake.hexdigest()


def hash_array_data(ar):
    # this function should stay consistent with all future versions
    # since this is the expensive part of the hashing
    if isinstance(ar, np.ndarray):
        ar = ar.ravel()
        if ar.dtype == np.object_:
            return {"type": "numpy", "data": str(uuid.uuid4()), "mask": None}
        if np.ma.isMaskedArray(ar):
            data_byte_ar = _to_bytes(ar.data)
            blake = blake3.blake3(data_byte_ar, multithreading=True)
            hash_data = {"type": "numpy", "data": blake.hexdigest(), "mask": None}
            if ar.mask is not True and ar.mask is not False and ar.mask is not np.True_ and ar.mask is not np.False_:
                mask_byte_ar = _to_bytes(ar.mask)
                blake = blake3.blake3(mask_byte_ar, multithreading=True)
                hash_data["mask"] = blake.hexdigest()
            return hash_data
        else:
            try:
                byte_ar = _to_bytes(ar)
            except ValueError:
                byte_ar = ar.copy().view(np.uint8)
            blake = blake3.blake3(byte_ar, multithreading=True)
            hash_data = {"type": "numpy", "data": blake.hexdigest(), "mask": None}
    else:
        if not isinstance(ar, pa.Array):
            try:
                ar = pa.array(ar)
            except Exception as e:
                raise ValueError(f'Cannot convert array {ar} to arrow array for hashing') from e
        blake = blake3.blake3(multithreading=True)
        buffer_hashes = []
        hash_data = {"type": "arrow", "buffers": buffer_hashes}
        for buffer in ar.buffers():
            if buffer is not None:
                # TODO: we need to make a copy here, a memoryview would be better
                # or possible patch the blake module to accept a memoryview https://github.com/oconnor663/blake3-py/issues/9
                # or feed in the buffer in batches
                # blake.update(buffer)
                blake.update(memoryview((buffer)).tobytes())
                buffer_hashes.append(blake.hexdigest())
            else:
                buffer_hashes.append(None)
    return hash_data


def hash_array(ar, hash_info=None, return_info=False):
    # this function can change over time, as it builds on top of the expensive part
    # (hash_array_data), so we can cheaply calculate new hashes if we pass on hash_info
    if hash_info is None:
        hash_info = hash_array_data(ar)
    if isinstance(ar, np.ndarray):
        if ar.dtype == np.object_:
            return hash_info['data']  # uuid, so always unique
        if np.ma.isMaskedArray(ar):
            if not (hash_info['type'] == 'numpy' and hash_info['data'] and hash_info['mask']):
                hash_info = hash_array_data(ar)
        else:
            if not (hash_info['type'] == 'numpy' and hash_info['data']):
                hash_info = hash_array_data(ar)
        keys = [HASH_VERSION, hash_info['type'], hash_info['data']]
        if hash_info['mask']:
            keys.append(hash_info['mask'])
    elif isinstance(ar, vaex.array_types.supported_arrow_array_types):
        if not (hash_info['type'] == 'arrow' and hash_info['buffers']):
            hash_info = hash_array_data(ar)
        keys = [HASH_VERSION]
        keys.extend(["NO_BUFFER" if not b else b for b in hash_info['buffers']])
    blake = blake3.blake3(multithreading=False)  # small amounts of data
    for key in keys:
        blake.update(key.encode('ascii'))
    hash = blake.hexdigest()
    if return_info:
        return hash, hash_info
    else:
        return hash


def to_supported_array(ar):
    if not isinstance(ar, supported_column_types):
        ar = np.asanyarray(ar)

    if isinstance(ar, np.ndarray) and ar.dtype.kind == 'U':
        ar = vaex.column.ColumnArrowLazyCast(ar, pa.string())
    elif isinstance(ar, np.ndarray) and ar.dtype.kind == 'O':
        ar_data = ar
        if np.ma.isMaskedArray(ar):
            ar_data = ar.data

        try:
            # "k != k" is a way to detect NaN's and NaT's
            types = list({type(k) for k in ar_data if k is not None and k == k})
        except ValueError:
            # If there is an array value in the column, Numpy throws a ValueError
            # "The truth value of an array with more than one element is ambiguous".
            # We don't handle this by default as it is a bit slower.
            def is_missing(k):
                if k is None:
                    return True
                try:
                    # a way to detect NaN's and NaT
                    return not (k == k)
                except ValueError:
                    # if a value is an array, this will fail, and it is a non-missing
                    return False
            types = list({type(k) for k in ar_data if k is not is_missing(k)})

        if len(types) == 1 and issubclass(types[0], str):
            # TODO: how do we know it should not be large_string?
            # self._dtypes_override[valid_name] = pa.string()
            ar = vaex.column.ColumnArrowLazyCast(ar, pa.string())
        if len(types) == 0:  # can only be if all nan right?
            ar = ar.astype(np.float64)
    return ar


def _concat_chunk_list(list_of_chunks):
    dict_of_list_of_arrays = collections.defaultdict(list)
    for chunks in list_of_chunks:
        for name, array in chunks.items():
            if isinstance(array, pa.ChunkedArray):
                dict_of_list_of_arrays[name].extend(array.chunks)
            else:
                dict_of_list_of_arrays[name].append(array)
    chunks = {name: vaex.array_types.concat(arrays) for name, arrays in dict_of_list_of_arrays.items()}
    return chunks


def _slice_of_chunks(chunks_ready_list, chunk_size):
    current_row_count = 0
    chunks_current_list = []
    while current_row_count < chunk_size and chunks_ready_list:
        chunks_current = chunks_ready_list.pop(0)
        chunk = list(chunks_current.values())[0]
        # chunks too large, split, and put back a part
        if current_row_count + len(chunk) > chunk_size:
            strict = True
            if strict:
                needed_length = chunk_size - current_row_count
                current_row_count += needed_length
                assert current_row_count == chunk_size


                chunks_head = {name: vaex.array_types.slice(chunk, 0, needed_length) for name, chunk in chunks_current.items()}
                chunks_current_list.append(chunks_head)
                chunks_extra = {name: vaex.array_types.slice(chunk, needed_length) for name, chunk in chunks_current.items()}
                chunks_ready_list.insert(0, chunks_extra)  # put back the extra in front
            else:
                current_row_count += len(chunk)
                chunks_current_list.append(chunks_current)
        else:
            current_row_count += len(chunk)
            chunks_current_list.append(chunks_current)
    return chunks_current_list, current_row_count

class Dataset(collections.abc.Mapping):
    def __init__(self):
        super().__init__()
        self._columns = frozendict()
        self._row_count = None

    @abstractmethod
    def _create_columns(self):
        pass

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['_columns']
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self._create_columns()

    def schema(self, array_type=None):
        return {name: vaex.array_types.data_type(col) for name, col in self.items()}

    def shapes(self):
        return {name: self.shape(name) for name, col in self.items()}

    def _set_row_count(self):
        if not self._columns:
            return
        values = list(self._columns.values())
        self._row_count = len(values[0])
        for name, value in list(self._columns.items())[1:]:
            if len(value) != self._row_count:
                raise ValueError(f'First columns has length {self._row_count}, while column {name} has length {len(value)}')

    @property
    def row_count(self):
        return self._row_count

    def project(self, *names):
        all = set(self)
        drop = all - set(names)
        return self.dropped(*list(drop))

    def concat(self, *others, resolver='flexible'):
        datasets = []
        if isinstance(self, DatasetConcatenated):
            datasets.extend(self.datasets)
        else:
            datasets.extend([self])
        for other in others:
            if isinstance(other, DatasetConcatenated):
                datasets.extend(other.datasets)
            else:
                datasets.extend([other])
        return DatasetConcatenated(datasets, resolver=resolver)

    def take(self, indices, masked=False):
        return DatasetTake(self, indices, masked=masked)

    def renamed(self, renaming):
        return DatasetRenamed(self, renaming)

    def merged(self, rhs):
        return DatasetMerged(self, rhs)

    def dropped(self, *names):
        return DatasetDropped(self, names)

    def __getitem__(self, item):
        if isinstance(item, slice):
            assert item.step in [1, None]
            return self.slice(item.start or 0, item.stop or self.row_count)
        return self._columns[item]

    def __len__(self):
        return len(self._columns)

    def __iter__(self):
        return iter(self._columns)

    def get_data(self, i1, i2, names):
        raise NotImplementedError

    def __eq__(self, rhs):
        if not isinstance(rhs, Dataset):
            return NotImplemented
        keys = set(self)
        keys_hashed = set(self._ids)
        missing = keys ^ keys_hashed
        if missing:
            raise ValueError(f'Comparing datasets where the left hand side is missing hashes for columns: {missing} (tip: use dataset.hashed())')
        keys = set(rhs)
        keys_hashed = set(rhs._ids)
        missing = keys ^ keys_hashed
        if missing:
            raise ValueError(f'Comparing datasets where the right hand side is missing hashes for columns: {missing} (tip: use dataset.hashed())')
        return self._ids == rhs._ids

    def __hash__(self):
        keys = set(self)
        keys_hashed = set(self._ids)
        missing = keys ^ keys_hashed
        if missing:
            raise ValueError(f'Trying to hash a dataset with unhashed columns: {missing} (tip: use dataset.hashed())')
        return hash(self._ids)

    def _default_lazy_chunk_iterator(self, array_map, columns, chunk_size, reverse=False):
        chunk_size = chunk_size or 1024**2
        chunk_count = (self.row_count + chunk_size - 1) // chunk_size
        chunks = range(chunk_count)
        if reverse:
            chunks = reversed(chunks)
        for i in chunks:
            i1 = i * chunk_size
            i2 = min((i + 1) * chunk_size, self.row_count)
            def reader(i1=i1, i2=i2):
                chunks = {k: array_map[k][i1:i2] for k in columns}
                length = i2 - i1
                for name, chunk in chunks.items():
                    assert len(chunk) == length, f'Oops, got a chunk ({name}) of length {len(chunk)} while it is expected to be of length {length} (at {i1}-{i2}'
                return chunks
            yield i1, i2, reader

    def _default_chunk_iterator(self, array_map, columns, chunk_size, reverse=False):
        for i1, i2, reader in self._default_lazy_chunk_iterator(array_map, columns, chunk_size, reverse):
            yield i1, i2, reader()

    @abstractmethod
    def chunk_iterator(self, columns, chunk_size=None, reverse=False):
        pass

    @abstractmethod
    def is_masked(self, column):
        pass

    @abstractmethod
    def shape(self, column):
        pass

    @abstractmethod
    def close(self):
        '''Close file handles or other resources, the DataFrame will not be in a usable state afterwards.'''
        pass

    @abstractmethod
    def slice(self, start, end):
        pass

    @abstractmethod
    def hashed(self):
        pass


class ColumnProxy(vaex.column.Column):
    '''To give the Dataset._columns object useful containers for debugging'''
    ds: Dataset

    def __init__(self, ds, name, type):
        self.ds = ds
        self.name = name
        self.dtype = type

    def __len__(self):
        return self.ds.row_count

    def to_numpy(self):
        return np.array(self)

    def __getitem__(self, item):
        if isinstance(item, slice):
            array_chunks = []
            ds = self.ds.__getitem__(item)
            for chunk_start, chunk_end, chunks in ds.chunk_iterator([self.name]):
                ar = chunks[self.name]
                if isinstance(ar, pa.ChunkedArray):
                    array_chunks.extend(ar.chunks)
                else:
                    array_chunks.append(ar)
            if len(array_chunks) == 1:
                return array_chunks[0]
            return vaex.array_types.concat(array_chunks)
        else:
            raise NotImplementedError


class DatasetRenamed(Dataset):
    def __init__(self, original, renaming):
        super().__init__()
        self.original = original
        self.renaming = renaming
        self.reverse = {v: k for k, v in renaming.items()}
        self._create_columns()
        self._ids = frozendict({renaming.get(name, name): ar for name, ar in original._ids.items()})
        self._set_row_count()

    def _create_columns(self):
        self._columns = frozendict({self.renaming.get(name, name): ar for name, ar in self.original.items()})

    def chunk_iterator(self, columns, chunk_size=None, reverse=False):
        columns = [self.reverse.get(name, name) for name in columns]
        for i1, i2, chunks in self.original.chunk_iterator(columns, chunk_size, reverse=reverse):
            yield i1, i2, {self.renaming.get(name, name): ar for name, ar in chunks.items()}

    def is_masked(self, column):
        return self.original.is_masked(self.reverse.get(column, column))

    def shape(self, column):
        return self.original.shape(self.reverse.get(column, column))

    def close(self):
        self.original.close()

    def slice(self, start, end):
        if start == 0 and end == self.row_count:
            return self
        return type(self)(self.original.slice(start, end), self.renaming)

    def hashed(self):
        if set(self._ids) == set(self):
            return self
        return type(self)(self.original.hashed(), self.renaming)


class DatasetConcatenated(Dataset):
    def __init__(self, datasets, resolver):
        self.datasets = datasets
        self.resolver = resolver
        if self.resolver == 'strict':
            for dataset in datasets[1:]:
                if set(dataset) != set(datasets[0]):
                    l = set(dataset)
                    r = set(datasets[0])
                    diff = l ^ r
                    raise NameError(f'Concatenating datasets with different names: {l} and {r} (difference: {diff})')
            self._schema = datasets[0].schema()
        elif self.resolver == 'flexible':
            schemas = [ds.schema() for ds in datasets]
            shapes = [ds.shapes() for ds in datasets]
            # try to keep the order of the original dataset
            schema_list_map = {}
            for schema in schemas:
                for name, type in schema.items():
                    if name not in schema_list_map:
                        schema_list_map[name] = []
            for name, type_list in schema_list_map.items():
                for schema in schemas:
                    # None means it is means the column is missing
                    type_list.append(schema.get(name))
            from .schema import resolver_flexible

            # shapes
            shape_list_map = {}
            for shape in shapes:
                for name, type in shape.items():
                    if name not in shape_list_map:
                        shape_list_map[name] = []
            for name, shape_list in shape_list_map.items():
                for shapes_ in shapes:
                    # None means it is means the column is missing
                    shape_list.append(shapes_.get(name))
            self._schema = {}
            self._shapes = {}

            for name in shape_list_map:
                self._schema[name], self._shapes[name] = resolver_flexible.resolve(schema_list_map[name], shape_list_map[name])
        else:
            raise ValueError(f'Invalid resolver {resolver}, choose between "strict" or "flexible"')

        self._create_columns()
        self._set_row_count()

    def _create_columns(self):
        columns = {}
        hashes = {}
        for name in self._schema:
            columns[name] = ColumnProxy(self, name, self._schema[name])
            if all(name in ds._ids for ds in self.datasets):
                hashes[name] = hash_combine(*[ds._ids[name] for ds in self.datasets])
        self._columns = frozendict(columns)
        self._ids = frozendict(hashes)

    def is_masked(self, column):
        return any(k.is_masked(column) for k in self.datasets)

    def shape(self, column):
        return self._shapes[column]

    def _set_row_count(self):
        self._row_count = sum(ds.row_count for ds in self.datasets)

    def schema(self, array_type=None):
        return self._schema.copy()

    def _chunk_iterator_non_strict(self, columns, chunk_size=None, reverse=False, start=0, end=None):
        end = self.row_count if end is None else end
        offset = 0
        for dataset in self.datasets:
            present = [k for k in columns if k in dataset]
            # skip over whole datasets
            if start >= offset + dataset.row_count:
                offset += dataset.row_count
                continue
            # we are past the end
            if end <= offset:
                break
            for i1, i2, chunks in dataset.chunk_iterator(present, chunk_size=chunk_size, reverse=reverse):
                # chunks = {name: vaex.array_types.to_arrow(ar) for name, ar in chunks.items()}
                length = i2 - i1
                chunk_start = offset
                chunk_end = offset + length
                if start >= chunk_end:  # we didn't find the beginning yet
                    offset += length
                    continue
                if end <= chunk_start:  # we are past the end
                    # assert False
                    break

                if start > chunk_start:
                    # this means we have to cut off a piece of the beginning
                    if end < chunk_end:
                        # AND the end
                        length = end - chunk_start  # without the start cut off
                        length -= start - chunk_start  # correcting for the start cut off
                        assert length > 0
                        chunks = {name: vaex.array_types.slice(ar, start - chunk_start, length) for name, ar in chunks.items()}
                        for name, ar in chunks.items():
                            assert len(ar) == length, f'Oops, array was expected to be of length {length} but was {len(ar)}'
                    else:
                        length -= start - chunk_start  # correcting for the start cut off
                        assert length > 0
                        chunks = {name: vaex.array_types.slice(ar, start - chunk_start) for name, ar in chunks.items()}
                        for name, ar in chunks.items():
                            assert len(ar) == length, f'Oops, array was expected to be of length {length} but was {len(ar)}'
                else:
                    if end < chunk_end:
                        # we only need to cut off a piece of the end
                        length = end - chunk_start
                        assert length > 0
                        chunks = {name: vaex.array_types.slice(ar, 0, length) for name, ar in chunks.items()}
                        for name, ar in chunks.items():
                            assert len(ar) == length, f'Oops, array was expected to be of length {length} but was {len(ar)}'

                from .schema import resolver_flexible
                allchunks = {name: resolver_flexible.align(length, chunks.get(name), self._schema[name], self._shapes[name]) for name in columns}
                yield {k: allchunks[k] for k in columns}
                offset += (i2 - i1)

    def chunk_iterator(self, columns, chunk_size=None, reverse=False, start=0, end=None):
        chunk_size = chunk_size or 1024*1024
        i1 = 0
        chunks_ready_list = []
        i1 = i2 = 0
        if not columns:
            end = self.row_count if end is None else end
            length = end - start
            i2 = min(length, i1 + chunk_size)
            while i1 < length:
                yield i1, i2, {}
                i1 = i2
                i2 = min(length, i1 + chunk_size)
            return

        for chunks in self._chunk_iterator_non_strict(columns, chunk_size, reverse=reverse, start=start, end=self.row_count if end is None else end):
            chunks_ready_list.append(chunks)
            total_row_count = sum([len(list(k.values())[0]) for k in chunks_ready_list])
            if total_row_count > chunk_size:
                chunks_current_list, current_row_count = vaex.dataset._slice_of_chunks(chunks_ready_list, chunk_size)
                i2 += current_row_count
                chunks = vaex.dataset._concat_chunk_list(chunks_current_list)
                yield i1, i2, chunks
                i1 = i2

        while chunks_ready_list:
            chunks_current_list, current_row_count = vaex.dataset._slice_of_chunks(chunks_ready_list, chunk_size)
            i2 += current_row_count
            chunks = vaex.dataset._concat_chunk_list(chunks_current_list)
            yield i1, i2, chunks
            i1 = i2

    def close(self):
        for ds in self.datasets:
            ds.close()

    def slice(self, start, end):
        if start == 0 and end == self.row_count:
            return self
        # TODO: we can be smarter here, and trim off some datasets
        return DatasetSliced(self, start=start, end=end)

    def hashed(self):
        if set(self._ids) == set(self):
            return self
        return type(self)([dataset.hashed() for dataset in self.datasets])


class DatasetTake(Dataset):
    def __init__(self, original, indices, masked):
        super().__init__()
        self.original = original
        self.indices = indices
        self.masked = masked
        self._hash_index = hash_array(indices)
        self._create_columns()
        self._set_row_count()

    def _create_columns(self):
        # if the columns in ds already have a ColumnIndex
        # we could do, direct_indices = df.column['bla'].indices[indices]
        # which should be shared among multiple ColumnIndex'es, so we store
        # them in this dict
        direct_indices_map = {}
        columns = {}
        hashes = {}
        for name, column in self.original.items():
            columns[name] = ColumnIndexed.index(column, self.indices, direct_indices_map, masked=self.masked)
            if name in self.original._ids:
                hashes[name] = hash_combine(self._hash_index, self.original._ids[name])
        self._columns = frozendict(columns)
        self._ids = frozendict(hashes)

    def is_masked(self, column):
        return self.original.is_masked(column)

    def shape(self, column):
        return self.original.shape(column)

    def chunk_iterator(self, columns, chunk_size=None, reverse=False):
        # TODO: we may be able to do this slightly more efficient by first
        # materializing the columns
        yield from self._default_chunk_iterator(self._columns, columns, chunk_size, reverse=reverse)

    def slice(self, start, end):
        if start == 0 and end == self.row_count:
            return self
        return DatasetSlicedArrays(self, start=start, end=end)

    def hashed(self):
        if set(self._ids) == set(self):
            return self
        return type(self)(self.original.hashed(), self.indices, self.masked)

    def close(self):
        self.original.close()


class DatasetSliced(Dataset):
    def __init__(self, original, start, end):
        super().__init__()
        self.original = original
        self.start = start
        self.end = end
        self._row_count = end - start
        self._create_columns()
        self._ids = {}

    def _create_columns(self):
        self._columns = {name: vaex.dataset.ColumnProxy(self, name, col.dtype) for name, col in self.original._columns.items()}

    def chunk_iterator(self, columns, chunk_size=None, reverse=False):
        yield from self.original.chunk_iterator(columns, chunk_size=chunk_size, reverse=reverse, start=self.start, end=self.end)

    def is_masked(self, column):
        return self.original.is_masked(column)

    def shape(self, column):
        return self.original.shape(column)

    def hashed(self):
        raise NotImplementedError

    def close(self):
        self.original.close()

    def slice(self, start, end):
        length = end - start
        start += self.start
        end = start + length
        if end > self.original.row_count:
            raise IndexError(f'Slice end ({end}) if larger than number of rows: {self.original.row_count}')
        return type(self)(self.original, start, end)


class DatasetSlicedArrays(Dataset):
    def __init__(self, original, start, end):
        super().__init__()
        # maybe we want to avoid slicing twice, and collapse it to 1?
        self.original = original
        self.start = start
        self.end = end
        # TODO: this is the old dataframe.trim method, we somehow need to test/capture that
        # if isinstance(column, array_types.supported_array_types):  # real array
        #     df.columns[name] = column[self._index_start:self._index_end]
        # else:
        #     df.columns[name] = column.trim(self._index_start, self._index_end)
        self._create_columns()
        self._ids = frozendict({name: hash_slice(hash, start, end) for name, hash in original._ids.items()})
        self._set_row_count()

    def _create_columns(self):
        columns = {}
        for name, column in self.original.items():
            if isinstance(column, array_types.supported_array_types):  # real array
                column = column[self.start:self.end]
            else:
                column = column.trim(self.start, self.end)
            columns[name] = column
        self._columns = frozendict(columns)

    def chunk_iterator(self, columns, chunk_size=None, reverse=False):
        yield from self._default_chunk_iterator(self._columns, columns, chunk_size, reverse=reverse)

    def is_masked(self, column):
        return self.original.is_masked(column)

    def shape(self, column):
        return self.original.shape(column)

    def hashed(self):
        if set(self._ids) == set(self):
            return self
        return type(self)(self.original.hashed(), self.start, self.end)

    def close(self):
        self.original.close()

    def slice(self, start, end):
        if start == 0 and end == self.row_count:
            return self
        length = end - start
        start += self.start
        end = start + length
        if end > self.original.row_count:
            raise IndexError(f'Slice end ({end}) if larger than number of rows: {self.original.row_count}')
        return type(self)(self.original, start, end)


class DatasetDropped(Dataset):
    def __init__(self, original, names):
        super().__init__()
        self.original = original
        self._dropped_names = tuple(names)
        self._create_columns()
        self._ids = frozendict({name: ar for name, ar in original._ids.items() if name not in names})
        self._set_row_count()

    def _create_columns(self):
        self._columns = frozendict({name: ar for name, ar in self.original.items() if name not in self._dropped_names})

    def chunk_iterator(self, columns, chunk_size=None, reverse=False):
        for column in columns:
            if column in self._dropped_names:
                raise KeyError(f'Oops, you tried to get column {column} while it is actually dropped')
        yield from self.original.chunk_iterator(columns, chunk_size=chunk_size, reverse=reverse)

    def is_masked(self, column):
        return self.original.is_masked(column)

    def shape(self, column):
        return self.original.shape(column)

    def hashed(self):
        if set(self._ids) == set(self):
            return self
        return type(self)(self.original.hashed(), self._dropped_names)

    def close(self):
        self.original.close()

    def slice(self, start, end):
        if start == 0 and end == self.row_count:
            return self
        return type(self)(self.original.slice(start, end), self._dropped_names)


class DatasetMerged(Dataset):
    def __init__(self, left, right):
        super().__init__()
        self.left = left
        self.right = right
        if self.left.row_count != self.right.row_count:
            raise ValueError(f'Merging datasets with unequal row counts ({self.left.row_count} != {self.right.row_count})')
        self._row_count = self.left.row_count
        overlap = set(left) & set(right)
        if overlap:
            raise NameError(f'Duplicate names: {overlap}')
        self._create_columns()
        self._ids = frozendict({**left._ids, **right._ids})
        self._set_row_count()

    def _create_columns(self):
        # TODO: for DatasetArray, we might want to just do this?
        # self._columns = frozendict({**left._columns, **right._columns})
        self._columns = {**{name: ColumnProxy(self.left, name, data_type(col)) for name, col in self.left._columns.items()},
                         **{name: ColumnProxy(self.right, name, data_type(col)) for name, col in self.right._columns.items()}}

    def chunk_iterator(self, columns, chunk_size=None, reverse=False):
        columns_left = [k for k in columns if k in self.left]
        columns_right = [k for k in columns if k in self.right]
        if not columns_left:
            yield from self.right.chunk_iterator(columns, chunk_size, reverse=reverse)
        elif not columns_right:
            yield from self.left.chunk_iterator(columns, chunk_size, reverse=reverse)
        else:
            for (i1, i2, ichunks), (j1, j2, jchunks) in zip(
                self.left.chunk_iterator(columns_left, chunk_size, reverse=reverse),
                self.right.chunk_iterator(columns_right, chunk_size, reverse=reverse)):
                # TODO: if one of the datasets does not respect the chunk_size (e.g. parquet)
                # this might fail
                assert i1 == j1
                assert i2 == j2
                yield i1, i2, {**ichunks, **jchunks}

    def is_masked(self, column):
        if column in self.left:
            return self.left.is_masked(column)
        else:
            return self.right.is_masked(column)

    def shape(self, column):
        if column in self.left:
            return self.left.shape(column)
        else:
            return self.right.shape(column)

    def hashed(self):
        if set(self._ids) == set(self):
            return self
        return type(self)(self.left.hashed(), self.right.hashed())

    def close(self):
        self.left.close()
        self.right.close()

    def slice(self, start, end):
        if start == 0 and end == self.row_count:
            return self
        return type(self)(self.left.slice(start, end), self.right.slice(start, end))


class DatasetArrays(Dataset):
    def __init__(self, mapping=None, **kwargs):
        super().__init__()
        if mapping is None:
            mapping = {}
        columns = {**mapping, **kwargs}
        columns = {key: to_supported_array(ar) for key, ar in columns.items()}
        # TODO: we finally want to get rid of datasets with no columns
        self._columns = frozendict(columns)
        self._ids = frozendict()
        self._set_row_count()

    def __getstate__(self):
        state = self.__dict__.copy()
        # here, we actually DO want to keep the columns
        # del state['_columns']
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        # self._create_columns()

    def _create_columns(self):
        pass

    def chunk_iterator(self, columns, chunk_size=None, reverse=False):
        yield from self._default_chunk_iterator(self._columns, columns, chunk_size, reverse=reverse)

    def is_masked(self, column):
        ar = self._columns[column]
        if not isinstance(ar, np.ndarray):
            ar = ar[0:1]  # take a small piece
        if isinstance(ar, np.ndarray):
            return np.ma.isMaskedArray(ar)
        else:
            return False  # an arrow array always has null value options

    def shape(self, column):
        ar = self._columns[column]
        if not isinstance(ar, np.ndarray):
            ar = ar[0:1]  # take a small piece
        if isinstance(ar, vaex.array_types.supported_arrow_array_types):
            return tuple()
        else:
            return ar.shape[1:]

    def merged(self, rhs):
        # TODO: if we don't allow emtpy datasets, we can remove this method
        if len(self) == 0:
            return rhs
        if len(rhs) == 0:
            return self
        # TODO: this is where we want to check if both are array like
        # and have faster version of merged
        return DatasetMerged(self, rhs)

    def slice(self, start, end):
        if start == 0 and end == self.row_count:
            return self
        return DatasetSlicedArrays(self, start=start, end=end)

    def hashed(self):
        if set(self._ids) == set(self):
            return self
        new = type(self)(self._columns)
        new._ids = frozendict({key: hash_array(array) for key, array in new._columns.items()})
        return new

    def close(self):
        pass  # nothing to do, maybe drop a refcount?

    # TODO: we might want to really get rid of these, since we want to avoid copying them over the network?
    # def dropped(self, names):

class DatasetFile(Dataset):
    """Datasets that map to a file can keep their ids/hashes in the file itself,
    or keep them in a meta file.
    """
    def __init__(self, path, write=False):
        super().__init__()
        self.path = path
        self.write = write
        self._columns = {}
        self._ids = {}
        self._frozen = False
        self._hash_calculations = 0  # track it for testing purposes
        self._hash_info = {}
        self._read_hashes()

    def _create_columns(self):
        pass

    @classmethod
    def quick_test(cls, path, fs_options={}, fs=None, *args, **kwargs):
        return False

    @classmethod
    def open(cls, path, *args, **kwargs):
        return cls(path, *args, **kwargs)

    def chunk_iterator(self, columns, chunk_size=None, reverse=False):
        yield from self._default_chunk_iterator(self._columns, columns, chunk_size, reverse=reverse)

    def is_masked(self, column):
        ar = self._columns[column]
        if not isinstance(ar, np.ndarray):
            ar = ar[0:1]  # take a small piece
        if isinstance(ar, np.ndarray):
            return np.ma.isMaskedArray(ar)
        else:
            return False  # an arrow array always has null value options

    def shape(self, column):
        ar = self._columns[column]
        if not isinstance(ar, np.ndarray):
            ar = ar[0:1]  # take a small piece
        if isinstance(ar, vaex.array_types.supported_arrow_array_types):
            return tuple()
        else:
            return ar.shape[1:]

    def slice(self, start, end):
        if start == 0 and end == self.row_count:
            return self
        return DatasetSlicedArrays(self, start=start, end=end)

    def _read_hashes(self):
        path_hashes = Path(self.path + '.d') / 'hashes.yaml'
        try:
            exists = path_hashes.exists()
        except OSError:  # happens for windows py<38
            exists = False
        if exists:
            with path_hashes.open() as f:
                hashes = vaex.utils.yaml_load(f)
                if hashes is None:
                    raise ValueError(f'{path_hashes} was probably truncated due to another process writing.')
                self._hash_info = hashes.get('columns', {})

    def _freeze(self):
        self._ids = frozendict(self._ids)
        self._columns = frozendict(self._columns)
        self._set_row_count()
        self._frozen = True

    def __getstate__(self):
        # we don't have the columns in the state, since we should be able
        # to get them from disk again
        return {
            'write': self.write,
            'path': self.path,
            '_ids': dict(self._ids)  # serialize the hases as non-frozen dict
        }

    def __setstate__(self, state):
        self.__dict__.update(state)
        # 'ctor' like initialization
        self._frozen = False
        self._hash_calculations = 0
        self._columns = {}
        self._hash_info = {}
        self._read_hashes()

    def add_column(self, name, data):
        self._columns[name] = data
        if self.write:
            return  # the columns don't include the final data
            # the hashes will be done in .freeze()
        hash_info = self._hash_info.get(name)
        if hash_info:
            hash, hash_info = hash_array(data, hash_info, return_info=True)
            self._ids[name] = hash
            self._hash_info[name] = hash_info  # always update the information

    @property
    def _local_hash_path(self):
        # TODO: support s3 and gcs
        # TODO: fallback directory when a user cannot write
        if Path(self.path).exists():
            directory = Path(self.path + '.d')
            directory.mkdir(exist_ok=True)
        else:
            o = urlparse(self.path)
            directory = Path(vaex.utils.get_private_dir('dataset', o.scheme, o.netloc, o.path[1:]))
        return directory / 'hashes.yaml'

    def hashed(self):
        if set(self._ids) == set(self):
            return self
        cls = type(self)
        # use pickle protocol to clone
        new = cls.__new__(cls)
        new.__setstate__(self.__getstate__())
        hashes = {}
        disk_cached_hashes = {}
        for name, column in new.items():
            hash_info = self._hash_info.get(name)
            if hash_info is None:
                logging.warning(f'Calculating hash for column {name} of length {len(column)} (1 time operation, will be cached on disk)')
                hash_info = hash_array_data(column)
            hash, hash_info = hash_array(column, hash_info, return_info=True)
            new._hash_calculations += 1
            hashes[name] = hash
            disk_cached_hashes[name] = hash_info
        new._ids = frozendict(hashes)
        new._hash_info = frozendict(disk_cached_hashes)
        path_hashes = new._local_hash_path
        # TODO: without this check, if multiple processes are writing (e.g. tests/execution_test.py::test_task_sum with ray)
        # this leads to a race condition, where we write the file, and while truncated, _read_hases() fails (because the file exists)
        # if new._hash_info != new._ids:
        if 1:  # TODO: file lock
            with path_hashes.open('w') as f:
                vaex.utils.yaml_dump(f, {'columns': dict(new._hash_info)})
        return new



from __future__ import absolute_import
import json
import numpy as np
import datetime
import re
import vaex.hash
import pyarrow as pa


serializers = []


def register(cls):
    serializers.append(cls)
    return cls


@register
class DateTime64Serializer:
    @staticmethod
    def can_encode(obj):
        return isinstance(obj, (np.datetime64, np.timedelta64))

    @staticmethod
    def encode(obj):
        value = int(obj.astype(int))
        typename = obj.dtype.name
        type, unit = re.match('(\w*)\[(\w*)\]', obj.dtype.name).groups()
        return {
            'type': type,
            'data': {
                'value': value,
                'unit': unit
            }
        }

    @staticmethod
    def can_decode(data):
        return data.get('type') in ['datetime64', 'timedelta64']

    @staticmethod
    def decode(data):
        dtype = np.dtype("%s[%s]" % (data['type'], data['data']['unit']))
        value = np.int64(data['data']['value']).astype(dtype)
        return value


@register
class NumpySerializer:
    @staticmethod
    def can_encode(obj):
        return isinstance(obj, np.ndarray)

    @staticmethod
    def encode(obj):
        if np.ma.isMaskedArray(obj):
            values = obj.data.tolist()
            mask = obj.mask.tolist()
        else:
            values = obj.tolist()
            mask = None
        dtype = str(obj.dtype)
        return {
            'type': 'ndarray',
            'data': {
                'values': values,
                'mask': mask,
                'dtype': dtype
            }
        }

    @staticmethod
    def can_decode(data):
        return data.get('type') == 'ndarray'

    @staticmethod
    def decode(data):
        dtype = np.dtype(data['data']['dtype'])
        if 'mask' in data['data'] and data['data']['mask'] is not None:
            value = np.ma.array(data['data']['values'], mask=data['data']['mask'], dtype=dtype)
        else:
            value = np.array(data['data']['values'], dtype)
        return value




@register
class ArrowSerializer:
    @staticmethod
    def can_encode(obj):
        return isinstance(obj, pa.Array)

    @staticmethod
    def encode(obj):
        encoding = vaex.encoding.Encoding()
        data = encoding.encode('arrow-array', obj)
        wiredata = vaex.encoding.inline.serialize(data, encoding)
        return {'type': 'arrow-array', 'data': wiredata}

    @staticmethod
    def can_decode(data):
        return data.get('type') == 'arrow-array'

    @staticmethod
    def decode(data):
        wiredata = data['data']
        encoding = vaex.encoding.Encoding()
        data = vaex.encoding.inline.deserialize(wiredata, encoding)
        ar = encoding.decode('arrow-array', data)
        return ar


@register
class OrdererSetSerializer:
    @staticmethod
    def can_encode(obj):
        return isinstance(obj, vaex.hash.ordered_set)

    @staticmethod
    def encode(obj):
        values = list(obj.extract().items())
        clsname = obj.__class__.__name__
        return {
            'type': clsname,
            'data': {
                'values': values,
                'count': obj.count,
                'nan_count': obj.nan_count,
                'missing_count': obj.null_count
            }
        }

    @staticmethod
    def can_decode(data):
        return data.get('type', '').startswith('ordered_set')

    @staticmethod
    def decode(data):
        clsname = data['type']
        cls = getattr(vaex.hash, clsname)
        value = cls(dict(data['data']['values']), data['data']['count'], data['data']['nan_count'], data['data']['missing_count'])
        return value


def encode(obj):
    for serializer in serializers:
        if serializer.can_encode(obj):
            return serializer.encode(obj)


class VaexJsonEncoder(json.JSONEncoder):
    def default(self, obj):
        encoded = encode(obj)
        if obj is not None and encoded is not None:
            return encoded
        if isinstance(obj, np.integer):
            return obj.item()
        elif isinstance(obj, np.floating):
            return obj.item()
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        elif isinstance(obj, np.bytes_):
            return obj.decode('UTF-8')
        elif isinstance(obj, bytes):
            return str(obj, encoding='utf-8');
        else:
            return super(VaexJsonEncoder, self).default(obj)

class VaexJsonDecoder(json.JSONDecoder):
    def __init__(self, *args, **kwargs):
        json.JSONDecoder.__init__(self, object_hook=self.object_hook, *args, **kwargs)
    
    def object_hook(self, dct):
        for serializer in serializers:
            if serializer.can_decode(dct):
                return serializer.decode(dct)
        return dct



import pkg_resources


packages = ['vaex', 'vaex-core', 'vaex-viz', 'vaex-hdf5', 'vaex-server', 'vaex-astro',
            'vaex-ui', 'vaex-jupyter', 'vaex-ml', 'vaex-arrow', 'vaex-graphql', 'vaex-enterprise']


def get_versions():
    def is_installed(p):
        try:
            pkg_resources.get_distribution(p)
            return True
        except pkg_resources.DistributionNotFound:
            return False
    return {p: pkg_resources.get_distribution(p).version for p in packages if is_installed(p)}



import os
import io
from .s3 import patch_profile

import pyarrow as pa
import pyarrow.fs

from . import split_options, FileProxy, split_scheme
from .cache import FileSystemHandlerCached
from ..cache import fingerprint


region_cache = {}
fs_arrow_cache = {}


def glob(path, fs_options={}):
    from .s3 import glob
    return glob(path, fs_options)


def parse(path, fs_options, for_arrow=False):
    # Remove this line for testing purposes to fake not having s3 support
    # raise pyarrow.lib.ArrowNotImplementedError('FOR TESTING')
    path, fs_options = split_options(path, fs_options)
    path = path.replace('arrow+s3://', 's3://')
    fullpath = path
    scheme, path = split_scheme(path)
    assert scheme == 's3'

    # anon is for backwards compatibility
    fs_options['anonymous'] = (fs_options.pop('anon', None) in [True, 'true', 'True', '1']) or (fs_options.pop('anonymous', None) in [True, 'true', 'True', '1'])
    fs_options = patch_profile(fs_options)
    use_cache = fs_options.pop('cache', 'true') in [True, 'true', 'True', '1']
    bucket = path.split('/')[0]
    if 'region' not in fs_options:
        # cache region
        if bucket not in region_cache:
            # we use this to get the default region
            file_system, _ = pa.fs.FileSystem.from_uri(fullpath)
            region = file_system.region
            region_cache[bucket] = region
        else:
            region = region_cache[bucket]
        fs_options['region'] = region
    # bucket and options make up a unique key
    key = fingerprint(bucket, fs_options)
    if key not in fs_arrow_cache:
        fs = pa.fs.S3FileSystem(**fs_options)
        fs_arrow_cache[key] = fs
    else:
        fs = fs_arrow_cache[key]
    if use_cache:
        fs = FileSystemHandlerCached(fs, scheme='s3', for_arrow=for_arrow)
        if for_arrow:
            fs = pyarrow.fs.PyFileSystem(fs)
    return fs, path



from __future__ import absolute_import
__author__ = 'maartenbreddels'
import time
import threading
import numpy as np

class Dummy(object):
	pass
progress = Dummy()
progress.value = 0
from bokeh.plotting import curdoc, figure
from functools import partial
def thread_safe_call(doc, f, *args, **kwargs):
	doc.add_next_tick_callback(partial(f, *args, **kwargs))

import logging
logger = logging.getLogger("vaex.ext.bokeh")

job = None
def debounced_threaded_update(bokeh_image, factory_rgba8, limits, delay=0.5):
	global job
	doc = curdoc()
	xlim, ylim = limits
	def factory(job):
		logger.debug("executing job")
		rgba8 = factory_rgba8()
		rgba32 = rgba8.view(dtype=np.uint32).reshape(rgba8.shape[:2])
		def update_bokeh():
			logger.debug("updating bokeh image")
			dw = [xlim[1] - xlim[0]]
			dh = [ylim[1] - ylim[0]]
			x = xlim[0]
			y = ylim[0]
			if 1:
				bokeh_image.data_source.data.update(dict(image=[rgba32], dw=dw, dh=dh, x=[x], y=[y]))
			else:
				bokeh_image.data_source.data["image"] = [rgba32]
				bokeh_image.data_source.data["x"] = x
				bokeh_image.data_source.data["y"] = y
				#bokeh_image.x = xlim[0]
				#bokeh_image.y = ylim[0]
				bokeh_image.data_source.data["dw"] = dw
				bokeh_image.data_source.data["dh"] = dh
			#image = p.image_rgba(image=[img], x=xlim[0], y=ylim[0], dw=[xlim[1]-xlim[0]], dh=[ylim[1]-ylim[0]])
		#thread_safe_call(doc, update_bokeh)
		doc.add_next_tick_callback(update_bokeh)
	if job:
		logger.debug("cancelling job")
		job.cancel()
	logger.debug("scheduling job")
	job = Job(factory, delay=delay)
	job.schedule()



def update(_, selection=False):
	global current_job
	doc = curdoc()
	def work(job, previous_job):
		#print "starting.."
		if previous_job:
			previous_job.cancel()
			previous_job.thread.join(5) # wait for 5 sec max
			#print "wait..."
		if job.cancelled:
			return
		def update_progress(fraction):
			def update_in_main_thread():
				progress.value = fraction/steps + step[0]/float(steps)
			thread_safe_call(doc, update_in_main_thread)
			#print progress.description, fraction, progress.value, job.cancelled#"\r",
			if fraction == 1:
				step[0] += 1
			return not job.cancelled
		try:
			import time
			callback = nyt.executor.signal_progress.connect(update_progress)
			if selection:
				steps = 1
				step = [0]
				progress.description = "Selecting"
				do_selection(job)
			progress.description = "Image"
			steps = 1
			step = [0]
			nyt.executor
			t0 = time.time()
			if not job.cancelled:
				rgba = create_image(job)
			if not job.cancelled:
				progress.description = "Done: %.1fs" % (time.time() - t0)
		except:
			if not job.cancelled:
				progress.description = "error"
			raise
		finally:
			nyt.executor.signal_progress.disconnect(callback)
		def update_widget_in_main_thread():
			url = vaex.image.rgba_to_url(rgba)
			im.src = url
			im.x = x_sc.min
			im.y = y_sc.max
			im.width = (x_sc.max - x_sc.min)
			im.height = -(y_sc.max - y_sc.min)
		if not job.cancelled:
			get_ioloop().add_callback(update_widget_in_main_thread)
	if current_job:
		current_job.cancel()
	current_job = Job(work, previous_job=current_job, delay=0.5)
	current_job.schedule()


import vaex.utils
import pyarrow.fs
s3fs = vaex.utils.optional_import('s3fs')

import vaex.file.cache
from . import split_options, split_scheme
from .cache import FileSystemHandlerCached
from .s3 import patch_profile
from ..cache import fingerprint


fs_cache = {}


def translate_options(fs_options):
    # translate options of arrow to s3fs
    fs_options = fs_options.copy()
    not_supported = {
        'role_arn', 'session_name', 'external_id', 'load_frequency', 'scheme', 'background_writes', 'profile', 'profile_name'
    }
    for key in not_supported:
        if key in fs_options:
             warnings.warn(f'The option {key} is not supported using s3fs instead of arrow, so it will be ignored')
             fs_options.pop(key)

    # top level
    mapping = {
        'anonymous': 'anon',
    }
    for key in list(fs_options):
        if key in mapping:
            fs_options[mapping[key]] = fs_options.pop(key)

    # client kwargs
    mapping = {
        'access_key': 'aws_access_key_id',
        'secret_key': 'aws_secret_access_key',
        'session_token': 'aws_session_token',
        'region': 'region_name',
        'endpoint_override': 'endpoint_url',
    }
    fs_options['client_kwargs'] = fs_options.get('client_kwargs', {})
    for key in list(fs_options):
        if key in mapping:
            fs_options['client_kwargs'][mapping[key]] = fs_options.pop(key)
    return fs_options


def glob(path, fs_options={}):
    if '?' in path:
        __, query = path[:path.index('?')], path[path.index('?'):]
    else:
        query = ''
    scheme, _ = split_scheme(path)
    path = path.replace('s3fs://', 's3://')
    path, fs_options = split_options(path, fs_options)
    # anon is for backwards compatibility
    if 'cache' in fs_options:
        del fs_options['cache']
    # standardize value, and make bool
    anon = (fs_options.pop('anon', None) in [True, 'true', 'True', '1']) or (fs_options.pop('anonymous', None) in [True, 'true', 'True', '1'])
    fs_options = patch_profile(fs_options)
    fs_options = translate_options(fs_options)
    s3 = s3fs.S3FileSystem(anon=anon, **fs_options)
    return [f'{scheme}://' + k + query for k in s3.glob(path)]


def parse(path, fs_options, for_arrow=False):
    path = path.replace('fsspec+s3://', 's3://')
    path, fs_options = split_options(path, fs_options)
    scheme, path = split_scheme(path)
    use_cache = fs_options.pop('cache', 'true') in [True, 'true', 'True', '1']
    # standardize value, and make bool
    anon = (fs_options.pop('anon', None) in [True, 'true', 'True', '1']) or (fs_options.pop('anonymous', None) in [True, 'true', 'True', '1'])
    fs_options = patch_profile(fs_options)
    fs_options = translate_options(fs_options)

    bucket = path.split('/')[0]
    key = fingerprint(bucket, fs_options)
    if key not in fs_cache:
        s3 = s3fs.S3FileSystem(anon=anon, default_fill_cache=False, **fs_options)
        fs_cache[key] = s3
    else:
        s3 = fs_cache[key]
    fs = pyarrow.fs.FSSpecHandler(s3)
    if use_cache:
        fs = FileSystemHandlerCached(fs, scheme='s3', for_arrow=for_arrow)
    if for_arrow:
        fs = pyarrow.fs.PyFileSystem(fs)
    return fs, path
